#!/usr/bin/env python
# Copyright 2017 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
PURPOSE:
    To generate .rpm and .deb packages and test their integrity.
USAGE:
    ./build.py [os_type;version0,...versionN;package_name]
    ./build.py [--options [opts]]
        Arguments:
            If non-flagged arguments are used, then the following format is
            expected:
                <os version>;[<list of versions>];<package name>
            example:
                'redhat';6,7;f5-sdk-1.2.3-1.el7.noarch.rpm
                This will generate a redhat rpm natively for redhat versions 6
                and 7 with the name 'f5-sdk-1.2.3-1.el7.noarch.rpm'
        Options:
            --os_type - required for the --os_version and specifies a
                particular OS type to be isntalled against [redhat|ubuntu]
            --os_version - specifies the version of the os_type to be created
                and tested for
        NOTE: these options must be accounted for in the
        ../Docker/<os_type>/<os_version> file and directory presence.
    If no options are supplied, then both .rpm and .deb packages are built
    automatically based upon how this script is set under the
    expected_operations variable.
EXPANDABILITY:
    In order to expand this to natively create a .deb or .rpm package on a new
    os_type/os_version combo, it is required to create the Docker directory's
    subsequent directories to match that which already exists.

    Then update the expected_operations varaible in this file to match that
    which already exists.

    Example for ubuntu 15.0's addition:
        mkdir ../Docker/ubuntu/15.0
        cp ../Docker/ubuntu/14.01/* ../Docker/ubuntu/15.0
        vim build.py  # change the expected_operations
            ...
            expected_operations = ... 'ubuntu', ['14.01', '15.0']...
            ...
"""

from __future__ import print_function

import errno
import fcntl
import getopt
import json
import os
import re
import sys
import traceback

from scripts import build_expectations
from scripts import configure
from scripts import construct_setups
from scripts import install_test
from scripts.terminal import terminal

from scripts.build_exceptions import BuildError
from scripts.build_exceptions import DebianError
from scripts.build_exceptions import RedhatError

from inspect import currentframe as cf
from inspect import getframeinfo as gfi


# Globals...
# Keeping from confusing namespace issues...
builds = build_expectations.Builds()
OperationSettings = builds.OperationSettings
expected_operations = builds.expected_operations


def load_json():
    """load_json

Loads the config.JSON file system object's contents and translate it into a
Python object.  This python object is then returned.
    """
    try:
        env = configure.get_env()
        configure.export_to_json(env)
    except SystemExit:
        raise BuildError(msg="configure failed!  Could not build targets!",
                         frame=gfi(cf()), errno=errno.ENOSYS)
    json_fl = os.path.dirname(sys.argv[0]) + "/scripts/config.JSON"
    data = _load_json(json_fl)
    return data


def _load_json(fl):
    data = {}
    if os.path.isfile(fl) and os.access(fl, os.R_OK):
        with open(fl, 'r') as fh:
            data = json.loads(fh.read())
    else:
        print("Unable to open JSON file: " + fl)
        reason = 'Cannot Access!' if os.path.isfile(fl) else \
            "File does not exist!"
        print(reason)
        raise BuildError(fl, reason, msg="Unable to open JSON file",
                         frame=gfi(cf()), errnum=errno.ENODATA)
    return data


def build_image_name(config, os_type, os_version):
    """build_image_name

    Returns a standard docker image name based upon package parameters
    """
    name = "%s%s-%s-builder" % (os_type, os_version, config['project'])
    return name


def _build_package(config, os_type, os_version):
    build_image = build_image_name(config, os_type, os_version)
    cmd = "docker build -t %s %s/Docker/%s/%s 2>&1" % \
        (build_image, config['dist_dir'], os_type, os_version)
    result = terminal([cmd], multi_stmt=True)
    if result.succeeded:
        cmd = 'docker run --rm -v %s:%s %s /var/wdir 2>&1' % \
            (config['working'], '/var/wdir', build_image)
        result = terminal([cmd], multi_stmt=True)
    return result


def build_debian(config, os_versions, os_type='ubuntu'):
    """build_debian

Builds for a specific debian operating system with os version
specified.  By default, it will use os_type='ubuntu'
    """
    def build_pkg(config, os_type, os_version):
        result = _build_package(config, os_type, os_version)
        if not result.succeeded:
            print(result.cli)
            raise DebianError(result, os_type, os_version, frame=gfi(cf()))

    error = 0
    if isinstance(os_versions, str):
        os_version = os_versions
        try:
            build_pkg(config, os_type, os_version)
        except DebianError as error:
            error.print_msg()
    else:
        for os_version in os_versions:
            try:
                build_pkg(config, os_type, os_version)
            except DebianError as error:
                error.print_msg()
    return error


def build_redhat(config, os_versions, os_type='redhat'):
    """build_redhat

Builds for a specific redhat operating system with os version specified. By
default, it will use os_type='redhat'
    """
    def build_pkg(config, os_type, os_version):
        result = _build_package(config, os_type, os_version)
        if not result.succeeded:
            print(result.cli)
            raise RedhatError(result, os_type, os_version, frame=gfi(cf()))

    error = 0
    if isinstance(os_versions, str):
        os_version = os_versions
        try:
            build_pkg(config, os_type, os_version)
        except RedhatError as error:
            error.print_msg()
    else:
        for os_version in os_versions:
            try:
                build_pkg(config, os_type, os_version)
            except RedhatError as error:
                error.print_msg()
    # this can be the last error to be reached!
    return error


def build_packages(config):
    """build_packages

Builds for the expected_operations' OperationSettings entries.

This function is intentionaly blind, dumb and stupid to produce what it can.
However, it should also exit with a non-zero when an error does occur to keep
from the caller thinking that things are 'okay' when they are not.
    """
    error = 0
    print("Building packages...")
    for operation in expected_operations:
        if operation.os_category == 'redhat':
            print("For Redhat...")
            function = build_redhat
        elif operation.os_category == 'debian':
            print("For Debian...")
            function = build_debian
        else:
            # can overwrite a previous error statement, but this does need to
            # be preserved...
            error = BuildError(operation, msg="Unexpected operation specified",
                               frame=gfi(cf()))
            function = Error.print_msg
        # preserve a previous iteration's error, if needed
        for os_version in operation.versions:
            error = function(config, os_version) if not error else error
            if error:
                break
        if error:
            print("A Failure occurred...", str(error))
            break
    if not error:
        print("Completed package builds...")
    return error


def _preliminary_construct(config):
    # this is an internal function; however, allows for quick and easy access
    # to execute these operations...
    error = 0
    construct_setups._construct_cfgs_from_json(config)
    try:
        pass
    except Exception as error:
        print("ASDASD")
        return error
    except SystemExit:
        error = BuildError(
            config['setup_cfg'],
            config['stdeb_cfg'],
            frame=gfi(cf()),
            errno=errno.ESPIPE,
            msg="Could not construct the setup files"
        )
    return error


def get_flagged_args():
    """get_flagged_args

Collects from the execution statement the arguments provided to this script.
The items are then interpretted and returned.  The object expected are the
KvP's:
    --os_type - the operating system type to be built
    --os_version - the operating system version to be built
NOTE: by not using these options, both Debian .deb and Redhat .rpm files are
generated for the operating systems and versions natively as set by a global
variable at the top of this script.
FURTHER NOTE: there should be a
    dist_dir/Docker/<os_type>/<os_version>/DockerFile*
Present for this script to work.
CONFIGURATION:  It is part of the standard for this script to run its own
configuration parameters to generate a:
    dist_dir/scripts/config.JSON
This is a part of a separate script and is executed by this one in an effort
to make this code as translatable from project segment to segment.
    """
    expected = ['os_type', 'os_version']
    arguments = {}
    try:
        opts, adds = \
            getopt.getopt(sys.argv, '', map(lambda x: x + "=", expected))
    except getopt.GetoptError as Error:
        print(str(Error))
        print("Defaulting to standard run...")
        return arguments
    for o, a in opts:
        opt = re.sub('^-+', '', o)
        if opt in expected:
            arguments[opt] = a
    if arguments:
        if 'os_type' not in arguments:
            print("Unsupported means of operation!")
            print("You can either specify both os_type and os_version " +
                  "or just os_type")
            arguments = {}
    return arguments


def get_args():
    global expected_operations
    new_list = list()
    arg_re = re.compile('([^;]+);(\S+)')
    breakout = {'redhat': 'redhat',
                'centos': 'redhat',
                'ubuntu': 'debian',
                'mint': 'debian'}
    for arg in sys.argv[1:]:
        match = arg_re.search(arg)
        if match:
            groups = match.groups()
            category = breakout[groups[0]] if groups[0] in breakout else \
                'redhat'
            new_list.append(OperationSettings(groups[0], category, groups[1]))
    if new_list:
        expected_operations = tuple(new_list)


def handle_flagged_args(args, config):
    """handle_flagged_args

This function will execute the appropriate steps for what arguments have been
given.  This is, essentially, a good "drop in" point for outside code to use
this module as long as args is a dict within the appropriate format or is an
empty dictionary.
    """
    error = None
    os_version = str()
    os_type = str()
    os_category = \
        BuildError(args, msg="Invalid build config", frame=gfi(cf()))
    # we're building for just one, and it should already be in
    # expected_operations...
    for item in expected_operations:
        if item.name == args['os_type']:
            for version in item.versions:
                os_version = version
                os_type = item.os_type
                if item.os_category == 'redhat':
                    os_category = build_redhat
                elif item.os_category == 'debian':
                    os_category = build_debian
                else:
                    error = False
                    os_category = os_category.print_msg
                break
    error = os_category(config, os_version, os_type=os_type) if error \
        else error


def store_json(obj, destination):
    """store_json

Takes in a json-portable object and a filesystem-based destination and stores
the json-portable object as JSON into the filesystem-based destination.

This is blind, dumb, and stupid; thus, it can fail if the object is more
complex than simple dict, list, int, str, etc. type object structures.
    """
    with open(destination, 'r+') as FH:
        fcntl.lockf(FH, fcntl.LOCK_EX)
        json_in = json.loads(FH.read())
        json_in.update(obj)  # obj overwrites items in json_in...
        FH.seek(0)
        FH.write(json.dumps(json_in, sort_keys=True, indent=4,
                            separators=(',', ': ')))


def test_packages(config):
    print("Initiating install tests...")
    test_object = install_test.InstallTest()
    test_object.execute_tests()
    if not test_object.failure_reason:
        print("Completed install tests")
    else:
        raise test_object.failure_reason


def main():
    """main

The standard entry point to this script.

This script will exit with a numeric error statement.  The following are
possible:
    errno.<POSIX> - Please look to the debugging messages and the meaning to
        The standard UNIX/LINUX POSIX meanings
    -1 - An unpredictable error has occurred.  Any number of problems may
        have happened; however, all of them were not predicted by the
        developer.  This is a bug that has been detected.
    0 - Execution, as far as the program is concerned, is complete and has
        been completed successfully.  Any errors that may have resulted,
        resulted without the program's knowledge and was the result of an
        error outside of the developer's ability to predict.
    """
    flagged_args = get_flagged_args()
    working = os.getcwd()
    config = load_json()
    errnum = 0
    error = _preliminary_construct(config)
    if not error and flagged_args:
        error = handle_flagged_args(flagged_args, config)
    elif not error:
        get_args()
        error = build_packages(config)
    error = test_packages(config)
    os.chdir(working)  # no matter where we were... return.
    if error:
        errnum = error.errnum
    return errnum


if __name__ == '__main__':
    errnum = 0
    try:
        errnum = main()
    except BuildError as Error:
        traceback.print_exc()
        print("Build was not successful at either a partial or total level.")
        errnum = errno.ESPIPE if not hasattr(Error, 'errnum') else \
            getattr(Error, 'errnum')
    except Exception as Error:
        traceback.print_exc()
        errnum = -1
        print("An unpredicted error has occurred!")
    else:
        if errnum:
            print("Please look above for error statements on what failed.")
    sys.exit(errnum)

# vim: set fileencoding=utf-8



# coding=utf-8
#
# Copyright 2020 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""BIG-IP® Network bwc module.

REST URI
    ``http://localhost/mgmt/tm/net/bwc``

GUI Path
    ``Acceleration --> Bandwidth Controllers``

REST Kind
    ``tm:net:bwc:*``
"""

from f5.bigip.resource import Collection
from f5.bigip.resource import OrganizingCollection
from f5.bigip.resource import Resource


class Bwc(OrganizingCollection):
    """BIG-IP® network bwc collection"""
    def __init__(self, net):
        super(Bwc, self).__init__(net)
        self._meta_data['allowed_lazy_attributes'] = [
            Policys,
        ]


class Policys(Collection):
    """BIG-IP® bwc policy sub-collection"""
    def __init__(self, bwc):
        super(Policys, self).__init__(bwc)
        self._meta_data['allowed_lazy_attributes'] = [Policy]
        self._meta_data['attribute_registry'] =\
            {'tm:net:bwc:policy:policystate': Policy}


class Policy(Resource):
    """BIG-IP® bwc policy sub-collection resource"""
    def __init__(self, policy_s):
        super(Policy, self).__init__(policy_s)
        self._meta_data['required_creation_parameters'].update(('partition',))
        self._meta_data['required_json_kind'] =\
            'tm:net:bwc:policy:policystate'



# coding=utf-8
#
# Copyright 2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""BIG-IP® Network route module.

REST URI
    ``http://localhost/mgmt/tm/net/route``

GUI Path
    ``Network --> Routes``

REST Kind
    ``tm:net:route:*``
"""

from f5.bigip.mixins import ExclusiveAttributesMixin
from f5.bigip.resource import Collection
from f5.bigip.resource import Resource
from f5.sdk_exception import MissingRequiredCreationParameter

from six import iterkeys


class Routes(Collection):
    """BIG-IP® network route collection"""
    def __init__(self, net):
        super(Routes, self).__init__(net)
        self._meta_data['allowed_lazy_attributes'] = [Route]
        self._meta_data['attribute_registry'] = {
            'tm:net:route:routestate': Route
        }


class Route(Resource, ExclusiveAttributesMixin):
    """BIG-IP® network route resource"""
    def __init__(self, route_s):
        super(Route, self).__init__(route_s)
        self._meta_data['required_json_kind'] = 'tm:net:route:routestate'
        self._meta_data['read_only_attributes'].append('network')
        self._meta_data['required_creation_parameters'].update(
            ('partition', 'name', 'network'))
        self._meta_data['exclusive_attributes'].append(
            ('blackhole', 'gw', 'tmInterface', 'pool'))

    def create(self, **kwargs):
        '''Create a Route on the BIG-IP® and the associated python object.

        One of the following gateways is required when creating the route
        objects: ``blackhole``, ``gw``, ``tmInterface``, ``pool``.

        :params kwargs: keyword arguments passed in from create call
        :raises: KindTypeMismatch
        :raises: MissingRequiredCreationParameter
        :raises: HTTPError
        :returns: Python Route object
        '''
        # We need to check that we have one of the available gateways set
        # when we create.  This isn't exactly the same as
        # required_creation_parameters because it needs to be one of the
        # gateways in the list.
        gateways = ['blackhole', 'gw', 'tmInterface', 'pool']
        kwkeys = list(iterkeys(kwargs))
        if not [k for k in kwkeys if k in gateways]:
            raise MissingRequiredCreationParameter(
                "One of %s gateways is required." % gateways
            )
        return self._create(**kwargs)



#!/usr/bin/env python
# Copyright 2017 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from collections import namedtuple

OperationSettings = namedtuple('OperationSettings',
                               'os_type, os_category, versions')


class Builds(object):
    """Builds

Creates an instance of a global object Builds.  This object can then be used to
extract globals.  This object cannot retain knowledge in regards to its
acknowledged attributes.

These globals are:
expected_operations - retains the build information on what will be built
OperationSettings - A namedtuple used throughout the project that keeps values
    that, if they were mutable, might otherwise be dangerous.
    """
    __OperationSettings = OperationSettings
    # When expanding the following variable, please make sure that you have a
    # Dockerfile that is found under the:
    # repo-dist/Docker/<os>/<os_version>/
    # Directory.  Without this, the script will fail.  The Dockerfile should
    # have its own ENTRYPOINT and handle the installation of the package on its
    # own.
    __expected_operations = \
        tuple([OperationSettings('ubuntu', 'debian', ['14.04']),
               OperationSettings('redhat', 'redhat', ['7'])])

    def __init__(self):
        pass

    @property
    def expected_operations(self):
        """expected_operations

An immutable, static tuple of OperationSettings instances used throughout the
build and test build code to determine what is built through automation.
        """
        return self.__expected_operations

    @property
    def OperationSettings(self):
        """OperationSettings

An immutable, static namedtuple that is used to construct the
expected_operations attribute.

This is also used throughout the project for expected_operations replacements
for its structure.
        """
        return self.__OperationSettings



from __future__ import absolute_import
# Copyright 2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from f5.devtools.source_engine import SourceEngine
from f5.devtools.template_engine import TemplateEngine

import os

UTILSDIR = os.path.abspath(os.path.dirname(__file__))
COMMONPREFIX = os.path.dirname(UTILSDIR)
SOURCE_ROOTDIR = os.path.join(COMMONPREFIX, 'f5', 'bigip')
TEMPLATEDIR = os.path.join(UTILSDIR, 'template_library')
DEVICECONFDIR = os.path.join(UTILSDIR, 'device_configs')


def main():
    temp_eng = TemplateEngine(TEMPLATEDIR, DEVICECONFDIR)
    for fname in os.listdir(DEVICECONFDIR):
        from_templ_src, uri = temp_eng.process_config_from_fname(fname)
        src_eng = SourceEngine(SOURCE_ROOTDIR, from_templ_src, uri)
        src_eng.integrate()

if __name__ == '__main__':
    main()



# coding=utf-8
#
#  Copyright 2014-2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""

REST URI
    ``http://localhost/mgmt/cm/cloud/tasks/configure-device-node``

REST Kind
    ``cm:cloud:tasks:configure-device-node:*``
"""

from f5.bigiq.resource import Collection
from f5.bigiq.resource import OrganizingCollection
from f5.bigiq.resource import TaskResource


class Tasks(OrganizingCollection):
    def __init__(self, cloud):
        super(Tasks, self).__init__(cloud)
        self._meta_data['object_has_stats'] = False
        self._meta_data['allowed_lazy_attributes'] = [
            Configure_Device_Nodes
        ]


class Configure_Device_Nodes(Collection):
    def __init__(self, tasks):
        super(Configure_Device_Nodes, self).__init__(tasks)
        self._meta_data['required_json_kind'] = \
            'cm:cloud:tasks:configure-device-node:configdevicenodetaskcollectionstate'  # NOQA
        self._meta_data['allowed_lazy_attributes'] = [Configure_Device_Node]
        self._meta_data['attribute_registry'] = {
            'cm:cloud:tasks:configure-device-node:configdevicenodetaskstate':
                Configure_Device_Node
        }


class Configure_Device_Node(TaskResource):
    def __init__(self, nodes):
        super(Configure_Device_Node, self).__init__(nodes)
        self._meta_data['required_json_kind'] = \
            'cm:cloud:tasks:configure-device-node:configdevicenodetaskstate'
        self._meta_data['required_load_parameters'] = {'id', }


class Reset_Devices(Collection):
    def __init__(self, tasks):
        super(Reset_Devices, self).__init__(tasks)
        self._meta_data['required_json_kind'] = \
            'cm:cloud:tasks:reset-device:resetdeviceconfigtaskcollectionstate'
        self._meta_data['allowed_lazy_attributes'] = [Reset_Device]
        self._meta_data['attribute_registry'] = {
            'cm:cloud:tasks:reset-device:resetdeviceconfigtaskstate':
                Reset_Device
        }


class Reset_Device(TaskResource):
    def __init__(self, resets):
        super(Reset_Device, self).__init__(resets)
        self._meta_data['required_json_kind'] = \
            'cm:cloud:tasks:reset-device:resetdeviceconfigtaskstate'
        self._meta_data['required_load_parameters'] = {'id', }
        self._meta_data['required_creation_parameters'] = {'deviceUUID', }



# Copyright 2015-2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from f5.bigiq import ManagementRoot
import pytest


@pytest.fixture(scope='module')
def mgmt_root(opt_bigip, opt_username, opt_password, opt_port, opt_token):
    return ManagementRoot(opt_bigip, opt_username, opt_password,
                          port=opt_port, auth_provider="local")



# coding=utf-8
#
# Copyright 2017 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from f5.bigip.resource import AsmResource
from f5.bigip.resource import Collection
from f5.sdk_exception import InvalidName


class Headers_s(Collection):
    """BIG-IP® ASM Headers sub-collection."""
    def __init__(self, policy):
        super(Headers_s, self).__init__(policy)
        self._meta_data['object_has_stats'] = False
        self._meta_data['allowed_lazy_attributes'] = [Header]
        self._meta_data['required_json_kind'] = 'tm:asm:policies:headers:headercollectionstate'
        self._meta_data['attribute_registry'] = {
            'tm:asm:policies:headers:headerstate': Header
        }


class Header(AsmResource):
    """BIG-IP® ASM Headers resource

    It should be noted that this Resource will cast everything to lowercase on the
    BIG-IP. Therefore, we require that you cast it before sending to the BIGIP.

    """
    def __init__(self, headers_s):
        super(Header, self).__init__(headers_s)
        self._meta_data['required_json_kind'] = \
            'tm:asm:policies:headers:headerstate'

    def create(self, **kwargs):
        if 'name' in 'kwargs':
            if kwargs['name'] != kwargs['name'].lower():
                raise InvalidName(
                    "The provided Header name must be lowercase"
                )
        return super(Header, self).create(**kwargs)

    def modify(self, **kwargs):
        if 'name' in 'kwargs':
            if kwargs['name'] != kwargs['name'].lower():
                raise InvalidName(
                    "The provided Header name must be lowercase"
                )
        return super(Header, self).modify(**kwargs)

    def update(self, **kwargs):
        if 'name' in 'kwargs':
            if kwargs['name'] != kwargs['name'].lower():
                raise InvalidName(
                    "The provided Header name must be lowercase"
                )
        return super(Header, self).update(**kwargs)

    def delete(self, **kwargs):
        if 'name' in 'kwargs':
            if kwargs['name'] != kwargs['name'].lower():
                raise InvalidName(
                    "The provided Header name must be lowercase"
                )
        return super(Header, self).delete(**kwargs)

    def load(self, **kwargs):
        if 'name' in 'kwargs':
            if kwargs['name'] != kwargs['name'].lower():
                raise InvalidName(
                    "The provided Header name must be lowercase"
                )
        return super(Header, self).load(**kwargs)



# coding=utf-8
#
# Copyright 2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""BIG-IP® system config module

REST URI
    ``http://localhost/mgmt/tm/sys/config``

GUI Path
    N/A

REST Kind
    ``tm:sys:config:*``
"""
from f5.bigip.mixins import CommandExecutionMixin
from f5.bigip.resource import UnnamedResource
from f5.sdk_exception import UnsupportedMethod


class Config(UnnamedResource,
             CommandExecutionMixin):
    def __init__(self, sys):
        super(Config, self).__init__(sys)
        self._meta_data['allowed_lazy_attributes'] = []
        self._meta_data['required_json_kind'] = \
            'tm:sys:config:configstate'
        self._meta_data['allowed_commands'].append('save')
        self._meta_data['allowed_commands'].append('load')

    def update(self, **kwargs):
        '''Update is not supported for Config

        :raises: UnsupportedOperation
        '''
        raise UnsupportedMethod(
            "%s does not support the update method" % self.__class__.__name__
        )

    def exec_cmd(self, command, **kwargs):
        """Normal save and load only need the command.

        To merge, just supply the merge and file arguments as kwargs like so:
        exec_cmd('load', merge=True, file='/path/to/file.txt')
        """

        if command == 'load':
            if kwargs:
                kwargs = dict(options=[kwargs])
        return self._exec_cmd(command, **kwargs)



# coding=utf-8
#
# Copyright 2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""iWorkflow® local cloud connector

REST URI
    ``http://localhost/mgmt/cm/cloud/connectors/local/``

REST Kind
    ``cm:cloud:connectors:*``
"""

from f5.iworkflow.resource import Collection
from f5.iworkflow.resource import Resource

# TODO(Implement device pointer)
# from f5.iworkflow.shared.resolver.device_groups.cm_cloud_managed_devices import Device  # NOQA


class Locals(Collection):
    def __init__(self, connectors):
        super(Locals, self).__init__(connectors)
        self._meta_data['allowed_lazy_attributes'] = [Local]
        self._meta_data['attribute_registry'] = {
            'cm:cloud:connectors:cloudconnectorstate': Local
        }


class Local(Resource):
    def __init__(self, locals):
        super(Local, self).__init__(locals)
        self._meta_data['required_json_kind'] = \
            'cm:cloud:connectors:cloudconnectorstate'
        self._meta_data['required_creation_parameters'] = {'name', }
        self._meta_data['allowed_lazy_attributes'] = [Nodes_s]
        self._meta_data['attribute_registry'] = {'': Nodes_s}


class Nodes_s(Collection):
    def __init__(self, local):
        super(Nodes_s, self).__init__(local)
        self._meta_data['required_json_kind'] = \
            'cm:cloud:nodes:nodecollection'
        self._meta_data['allowed_lazy_attributes'] = [Node]
        self._meta_data['attribute_registry'] = {
            'cm:cloud:nodes:node': Node
        }


class Node(Resource):
    def __init__(self, nodes):
        super(Node, self).__init__(nodes)
        self._meta_data['required_creation_parameters'] = set()
        self._meta_data['required_load_parameters'] = {'uuid', }
        self._meta_data['required_json_kind'] = \
            'cm:cloud:nodes:node'



# coding=utf-8
#
# Copyright 2014-2015 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""BIG-IP® Local Traffic Manager (LTM) rule module.

REST URI
    ``http://localhost/mgmt/tm/ltm/rule``

GUI Path
    ``Local Traffic --> Rules``

REST Kind
    ``tm:ltm:rule:*``
"""

from f5.bigip.resource import Collection
from f5.bigip.resource import Resource


class Rules(Collection):
    """BIG-IP® LTM rule collection"""
    def __init__(self, ltm):
        super(Rules, self).__init__(ltm)
        self._meta_data['allowed_lazy_attributes'] = [Rule]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:rule:rulestate': Rule}


class Rule(Resource):
    """BIG-IP® LTM rule resource"""
    def __init__(self, rule_s):
        super(Rule, self).__init__(rule_s)
        self._meta_data['required_json_kind'] = 'tm:ltm:rule:rulestate'
        self._meta_data['required_creation_parameters'].update(
            ('name', 'partition', 'apiAnonymous'))



# coding=utf-8
#
# Copyright 2015-2017 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""BIG-IP® Advanced Firewall Manager™ (AFM®) module.

REST URI
    ``http://localhost/mgmt/tm/security/analytics``

GUI Path
    ``Security --> Network Firewall``

REST Kind
    ``tm:security:analytics:*``
"""
from f5.bigip.resource import OrganizingCollection
from f5.bigip.resource import UnnamedResource


class Analytics(OrganizingCollection):
    """BIG-IP® AFM® Analytics organizing collection."""

    def __init__(self, security):
        super(Analytics, self).__init__(security)
        self._meta_data['allowed_lazy_attributes'] = [Settings]


class Settings(UnnamedResource):
    """BIG-IP® Analytics settings resource"""
    def __init__(self, settings):
        super(Settings, self).__init__(settings)
        self._meta_data['required_json_kind'] = \
            'tm:security:analytics:settings:settingsstate'



# coding=utf-8
#
#  Copyright 2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""iWorkflow® iApp templates

REST URI
    ``http://localhost/mgmt/cm/cloud/provider/templates/iapp``

REST Kind
    ``cm:cloud:provider:templates:iapp*``
"""

from f5.iworkflow.resource import Collection
from f5.iworkflow.resource import Resource


class Iapps(Collection):
    def __init__(self, templates):
        super(Iapps, self).__init__(templates)
        self._meta_data['required_json_kind'] = \
            'cm:cloud:templates:iapp:iapptemplatecollectionworkerstate'
        self._meta_data['allowed_lazy_attributes'] = [
            Iapp
        ]
        self._meta_data['attribute_registry'] = {
            'cm:cloud:templates:iapp:iapptemplateitemstate': Iapp
        }


class Iapp(Resource):
    def __init__(self, iapps):
        super(Iapp, self).__init__(iapps)
        self._meta_data['required_json_kind'] = \
            'cm:cloud:templates:iapp:iapptemplateitemstate'
        self._meta_data['required_load_parameters'] = {'name', }



# coding=utf-8
#
# Copyright 2017 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""BIG-IP® auth module

REST URI
    ``http://localhost/mgmt/tm/auth/radius-server``

GUI Path
    ``System --> Users --> Authentication``

REST Kind
    ``tm:auth:radius-server:*``
"""

from f5.bigip.resource import Collection
from f5.bigip.resource import Resource


class Radius_Servers(Collection):
    """BIG-IP® tacacs server collection"""
    def __init__(self, auth):
        super(Radius_Servers, self).__init__(auth)
        self._meta_data['allowed_lazy_attributes'] = [Radius_Server]
        self._meta_data['attribute_registry'] = \
            {'tm:auth:radius-server:radius-serverstate': Radius_Server}


class Radius_Server(Resource):
    """BIG-IP® tacacs server resource"""
    def __init__(self, radius_servers):
        super(Radius_Server, self).__init__(radius_servers)
        self._meta_data['required_json_kind'] = \
            'tm:auth:radius-server:radius-serverstate'



# coding=utf-8
#
# Copyright 2015 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""BIG-IP® System (sys) module

REST URI
    ``http://localhost/mgmt/tm/sys/``

GUI Path
    ``System``

REST Kind
    ``tm:sys:*``
"""

from f5.bigip.resource import OrganizingCollection
from f5.bigip.tm.sys.application import Application
from f5.bigip.tm.sys.config import Config
from f5.bigip.tm.sys.crypto import Crypto
from f5.bigip.tm.sys.daemon_log_settings import Daemon_Log_Settings
from f5.bigip.tm.sys.db import Dbs
from f5.bigip.tm.sys.dns import Dns
from f5.bigip.tm.sys.failover import Failover
from f5.bigip.tm.sys.file import File
from f5.bigip.tm.sys.folder import Folders
from f5.bigip.tm.sys.global_settings import Global_Settings
from f5.bigip.tm.sys.httpd import Httpd
from f5.bigip.tm.sys.icall import Icall
from f5.bigip.tm.sys.log_config import Log_Config
from f5.bigip.tm.sys.management_ip import Management_Ips
from f5.bigip.tm.sys.management_route import Management_Routes
from f5.bigip.tm.sys.ntp import Ntp
from f5.bigip.tm.sys.performance import Performances
from f5.bigip.tm.sys.provision import Provision
from f5.bigip.tm.sys.sflow import Sflow
from f5.bigip.tm.sys.smtp_server import Smtp_Servers
from f5.bigip.tm.sys.snmp import Snmp
from f5.bigip.tm.sys.software import Software
from f5.bigip.tm.sys.sshd import Sshd
from f5.bigip.tm.sys.syslog import Syslog
from f5.bigip.tm.sys.ucs import Ucs
from f5.bigip.tm.sys.version import Version


class Sys(OrganizingCollection):
    """BIG-IP® System (sys) organizing collection."""
    def __init__(self, tm):
        super(Sys, self).__init__(tm)
        self._meta_data['allowed_lazy_attributes'] = [
            Application,
            Config,
            Crypto,
            Daemon_Log_Settings,
            Dbs,
            Dns,
            Failover,
            File,
            Folders,
            Global_Settings,
            Httpd,
            Icall,
            Log_Config,
            Management_Ips,
            Management_Routes,
            Ntp,
            Performances,
            Provision,
            Sflow,
            Smtp_Servers,
            Snmp,
            Software,
            Sshd,
            Syslog,
            Ucs,
            Version
        ]



# coding=utf-8
#
# Copyright 2015-2017 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""BIG-IP® Advanced Firewall Manager™ (AFM®) module.

REST URI
    ``http://localhost/mgmt/tm/security/dos``

GUI Path
    ``Security --> Dos Protection``

REST Kind
    ``tm:security:dos*``
"""

from f5.bigip.mixins import CheckExistenceMixin
from f5.bigip.resource import Collection
from f5.bigip.resource import OrganizingCollection
from f5.bigip.resource import Resource
from f5.sdk_exception import NonExtantApplication

from distutils.version import LooseVersion


class Dos(OrganizingCollection):
    """BIG-IP® DOS organizing collection."""

    def __init__(self, security):
        super(Dos, self).__init__(security)
        self._meta_data['allowed_lazy_attributes'] = [Profiles]


class Profiles(Collection):
    """BIG-IP® Dos Profile collection"""
    def __init__(self, dos):
        super(Profiles, self).__init__(dos)
        self._meta_data['allowed_lazy_attributes'] = [Profile]
        self._meta_data['attribute_registry'] = \
            {'tm:security:dos:profile:profilestate': Profile}


class Profile(Resource):
    """BIG-IP® Dos Profile resource"""
    def __init__(self, profile_s):
        super(Profile, self).__init__(profile_s)
        self._meta_data['required_json_kind'] = \
            'tm:security:dos:profile:profilestate'
        self._meta_data['required_creation_parameters'].update(('partition',))
        self._meta_data['attribute_registry'] = \
            {'tm:security:dos:profile:application:applicationcollectionstate':
                Applications,
             'tm:security:dos:profile:dos-network:dos-networkcollectionstate':
                 Dos_Networks,
             'tm:security:dos:profile:protocol-dns:'
             'protocol-dnscollectionstate': Protocol_Dns_s,
             'tm:security:dos:profile:protocol-sip:protocol'
             '-sipcollectionstate': Protocol_Sips}


class Applications(Collection):
    """BIG-IP® Dos Profile Application sub-collection"""
    def __init__(self, profile):
        super(Applications, self).__init__(profile)
        self._meta_data['required_json_kind'] = \
            'tm:security:dos:profile:application:applicationcollectionstate'
        self._meta_data['allowed_lazy_attributes'] = [Application]
        self._meta_data['attribute_registry'] = \
            {'tm:security:dos:profile:application:applicationstate':
                Application}


class Application(Resource, CheckExistenceMixin):
    """BIG-IP® Dos Profile Application resource"""
    def __init__(self, applications):
        super(Application, self).__init__(applications)
        self._meta_data['required_json_kind'] = \
            'tm:security:dos:profile:application:applicationstate'
        self.tmos_ver = self._meta_data['bigip']._meta_data['tmos_version']

    def load(self, **kwargs):
        """Custom load method to address issue in 11.6.0 Final,

        where non existing objects would be True.
        """
        if LooseVersion(self.tmos_ver) == LooseVersion('11.6.0'):
            return self._load_11_6(**kwargs)
        else:
            return super(Application, self)._load(**kwargs)

    def _load_11_6(self, **kwargs):
        """Must check if rule actually exists before proceeding with load."""
        if self._check_existence_by_collection(self._meta_data['container'],
                                               kwargs['name']):
            return super(Application, self)._load(**kwargs)
        msg = 'The application resource named, {}, does not exist on the ' \
              'device.'.format(kwargs['name'])
        raise NonExtantApplication(msg)

    def exists(self, **kwargs):
        """Some objects when deleted still return when called by their

        direct URI, this is a known issue in 11.6.0.
        """

        if LooseVersion(self.tmos_ver) == LooseVersion('11.6.0'):
            return self._exists_11_6(**kwargs)
        else:
            return super(Application, self)._load(**kwargs)

    def _exists_11_6(self, **kwargs):
        """Check rule existence on device."""

        return self._check_existence_by_collection(
            self._meta_data['container'], kwargs['name'])


class Dos_Networks(Collection):
    """BIG-IP® Dos Profile Dos Networks sub-collection"""
    def __init__(self, profile):
        super(Dos_Networks, self).__init__(profile)
        self._meta_data['required_json_kind'] = \
            'tm:security:dos:profile:dos-network:dos-networkcollectionstate'
        self._meta_data['allowed_lazy_attributes'] = [Dos_Network]
        self._meta_data['attribute_registry'] = \
            {'tm:security:dos:profile:dos-network:dos-networkstate':
                Dos_Network}


class Dos_Network(Resource, CheckExistenceMixin):
    """BIG-IP® Dos Profile Dos Network resource"""
    def __init__(self, dos_networks):
        super(Dos_Network, self).__init__(dos_networks)
        self._meta_data['required_json_kind'] = \
            'tm:security:dos:profile:dos-network:dos-networkstate'
        self.tmos_ver = self._meta_data['bigip']._meta_data['tmos_version']

    def load(self, **kwargs):
        """Custom load method to address issue in 11.6.0 Final,

        where non existing objects would be True.
        """
        if LooseVersion(self.tmos_ver) == LooseVersion('11.6.0'):
            return self._load_11_6(**kwargs)
        else:
            return super(Dos_Network, self)._load(**kwargs)

    def _load_11_6(self, **kwargs):
        """Must check if rule actually exists before proceeding with load."""
        if self._check_existence_by_collection(self._meta_data['container'],
                                               kwargs['name']):
            return super(Dos_Network, self)._load(**kwargs)
        msg = 'The application resource named, {}, does not exist on the ' \
              'device.'.format(kwargs['name'])
        raise NonExtantApplication(msg)

    def exists(self, **kwargs):
        """Some objects when deleted still return when called by their

        direct URI, this is a known issue in 11.6.0.
        """

        if LooseVersion(self.tmos_ver) == LooseVersion('11.6.0'):
            return self._exists_11_6(**kwargs)
        else:
            return super(Dos_Network, self)._load(**kwargs)

    def _exists_11_6(self, **kwargs):
        """Check rule existence on device."""

        return self._check_existence_by_collection(
            self._meta_data['container'], kwargs['name'])


class Protocol_Dns_s(Collection):
    """BIG-IP® Dos Profile Protocol Dns sub-collection"""
    def __init__(self, profile):
        super(Protocol_Dns_s, self).__init__(profile)
        self._meta_data['required_json_kind'] = \
            'tm:security:dos:profile:protocol-dns:protocol-dnscollectionstate'
        self._meta_data['allowed_lazy_attributes'] = [Protocol_Dns]
        self._meta_data['attribute_registry'] = \
            {'tm:security:dos:profile:protocol-dns:protocol-dnsstate':
                Protocol_Dns}


class Protocol_Dns(Resource, CheckExistenceMixin):
    """BIG-IP® Dos Profile Protocol Dns resource"""
    def __init__(self, protocol_dns_s):
        super(Protocol_Dns, self).__init__(protocol_dns_s)
        self._meta_data['required_json_kind'] = \
            'tm:security:dos:profile:protocol-dns:protocol-dnsstate'
        self.tmos_ver = self._meta_data['bigip']._meta_data['tmos_version']

    def load(self, **kwargs):
        """Custom load method to address issue in 11.6.0 Final,

        where non existing objects would be True.
        """
        if LooseVersion(self.tmos_ver) == LooseVersion('11.6.0'):
            return self._load_11_6(**kwargs)
        else:
            return super(Protocol_Dns, self)._load(**kwargs)

    def _load_11_6(self, **kwargs):
        """Must check if rule actually exists before proceeding with load."""
        if self._check_existence_by_collection(self._meta_data['container'],
                                               kwargs['name']):
            return super(Protocol_Dns, self)._load(**kwargs)
        msg = 'The application resource named, {}, does not exist on the ' \
              'device.'.format(kwargs['name'])
        raise NonExtantApplication(msg)

    def exists(self, **kwargs):
        """Some objects when deleted still return when called by their

        direct URI, this is a known issue in 11.6.0.
        """

        if LooseVersion(self.tmos_ver) == LooseVersion('11.6.0'):
            return self._exists_11_6(**kwargs)
        else:
            return super(Protocol_Dns, self)._load(**kwargs)

    def _exists_11_6(self, **kwargs):
        """Check rule existence on device."""

        return self._check_existence_by_collection(
            self._meta_data['container'], kwargs['name'])


class Protocol_Sips(Collection):
    """BIG-IP® Dos Profile Protocol Sip sub-collection"""
    def __init__(self, profile):
        super(Protocol_Sips, self).__init__(profile)
        self._meta_data['required_json_kind'] = \
            'tm:security:dos:profile:protocol-sip:protocol-sipcollectionstate'
        self._meta_data['allowed_lazy_attributes'] = [Protocol_Sip]
        self._meta_data['attribute_registry'] = \
            {'tm:security:dos:profile:protocol-sip:protocol-sipstate':
                Protocol_Sip}


class Protocol_Sip(Resource, CheckExistenceMixin):
    """BIG-IP® Dos Profile Protocol Sip resource"""
    def __init__(self, protocol_sips):
        super(Protocol_Sip, self).__init__(protocol_sips)
        self._meta_data['required_json_kind'] = \
            'tm:security:dos:profile:protocol-sip:protocol-sipstate'
        self.tmos_ver = self._meta_data['bigip']._meta_data['tmos_version']

    def load(self, **kwargs):
        """Custom load method to address issue in 11.6.0 Final,

        where non existing objects would be True.
        """
        if LooseVersion(self.tmos_ver) == LooseVersion('11.6.0'):
            return self._load_11_6(**kwargs)
        else:
            return super(Protocol_Sip, self)._load(**kwargs)

    def _load_11_6(self, **kwargs):
        """Must check if rule actually exists before proceeding with load."""
        if self._check_existence_by_collection(self._meta_data['container'],
                                               kwargs['name']):
            return super(Protocol_Sip, self)._load(**kwargs)
        msg = 'The application resource named, {}, does not exist on the ' \
              'device.'.format(kwargs['name'])
        raise NonExtantApplication(msg)

    def exists(self, **kwargs):
        """Some objects when deleted still return when called by their

        direct URI, this is a known issue in 11.6.0.
        """

        if LooseVersion(self.tmos_ver) == LooseVersion('11.6.0'):
            return self._exists_11_6(**kwargs)
        else:
            return super(Protocol_Sip, self)._load(**kwargs)

    def _exists_11_6(self, **kwargs):
        """Check rule existence on device."""

        return self._check_existence_by_collection(
            self._meta_data['container'], kwargs['name'])



# coding=utf-8
#
#  Copyright 2014-2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""

REST URI
    ``http://localhost/mgmt/cm/asm/tasks/declare-mgmt-authority``

REST Kind
    ``cm:device:licensing:pool:regkey:licenses:*``
"""

from f5.bigiq.resource import Collection
from f5.bigiq.resource import OrganizingCollection
from f5.bigiq.resource import TaskResource


class Tasks(OrganizingCollection):
    def __init__(self, asm):
        super(Tasks, self).__init__(asm)
        self._meta_data['object_has_stats'] = False
        self._meta_data['allowed_lazy_attributes'] = [
            Declare_Mgmt_Authority_s
        ]


class Declare_Mgmt_Authority_s(Collection):
    def __init__(self, tasks):
        super(Declare_Mgmt_Authority_s, self).__init__(tasks)
        self._meta_data['required_json_kind'] = \
            'cm:asm:tasks:declare-mgmt-authority:dmataskcollectionstate'  # NOQA
        self._meta_data['allowed_lazy_attributes'] = [Declare_Mgmt_Authority]
        self._meta_data['attribute_registry'] = {
            'cm:asm:tasks:declare-mgmt-authority:dmataskitemstate':
                Declare_Mgmt_Authority
        }


class Declare_Mgmt_Authority(TaskResource):
    def __init__(self, authorities):
        super(Declare_Mgmt_Authority, self).__init__(authorities)
        self._meta_data['required_json_kind'] = \
            'cm:asm:tasks:declare-mgmt-authority:dmataskitemstate'
        self._meta_data['required_load_parameters'] = {'id', }



# coding=utf-8
#
# Copyright 2017 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
BIG-IP® system daemon log settings module

REST URI
    ``http://localhost/mgmt/tm/sys/daemon-log-settings``

tmsh Path
    ``sys --> daemon-log-settings --> all-properties``

GUI Path
    ``system --> logs --> configuration --> options``

REST Kind
    ``tm:sys:daemon-log-settings:*``
"""

from f5.bigip.resource import OrganizingCollection
from f5.bigip.resource import UnnamedResource


class Daemon_Log_Settings(OrganizingCollection):
    def __init__(self, sys):
        super(Daemon_Log_Settings, self).__init__(sys)
        self._meta_data['allowed_lazy_attributes'] = [
            Clusterd,
            Csyncd,
            Icrd,
            Lind,
            Mcpd,
            Tmm
        ]


class Clusterd(UnnamedResource):
    """BIG-IP® system daemon log settings clusterd resource

    The object only supports load and update because it is an
    unnamed resource.

    .. note::

        This is an unnamed resource so it has not ~Partition~Name pattern
        at the end of its URI.
    """
    def __init__(self, daemon_log_settings):
        super(Clusterd, self).__init__(daemon_log_settings)
        self._meta_data['required_load_parameters'] = set()
        self._meta_data['required_json_kind'] =\
            'tm:sys:daemon-log-settings:clusterd:clusterdstate'


class Csyncd(UnnamedResource):
    """BIG-IP® system daemon log settings csyncd resource

    The object only supports load and update because it is an
    unnamed resource.

    .. note::

        This is an unnamed resource so it has not ~Partition~Name pattern
        at the end of its URI.
    """
    def __init__(self, daemon_log_settings):
        super(Csyncd, self).__init__(daemon_log_settings)
        self._meta_data['required_load_parameters'] = set()
        self._meta_data['required_json_kind'] =\
            'tm:sys:daemon-log-settings:csyncd:csyncdstate'


class Icrd(UnnamedResource):
    """BIG-IP® system daemon log settings icrd resource

    The object only supports load and update because it is an
    unnamed resource.

    .. note::

        This is an unnamed resource so it has not ~Partition~Name pattern
        at the end of its URI.
    """
    def __init__(self, daemon_log_settings):
        super(Icrd, self).__init__(daemon_log_settings)
        self._meta_data['required_load_parameters'] = set()
        self._meta_data['required_json_kind'] =\
            'tm:sys:daemon-log-settings:icrd:icrdstate'


class Lind(UnnamedResource):
    """BIG-IP® system daemon log settings lind resource

    The object only supports load and update because it is an
    unnamed resource.

    .. note::

        This is an unnamed resource so it has not ~Partition~Name pattern
        at the end of its URI.
    """
    def __init__(self, daemon_log_settings):
        super(Lind, self).__init__(daemon_log_settings)
        self._meta_data['required_load_parameters'] = set()
        self._meta_data['required_json_kind'] =\
            'tm:sys:daemon-log-settings:lind:lindstate'


class Mcpd(UnnamedResource):
    """BIG-IP® system daemon log settings mcpd resource

    The object only supports load and update because it is an
    unnamed resource.

    .. note::

        This is an unnamed resource so it has not ~Partition~Name pattern
        at the end of its URI.
    """
    def __init__(self, daemon_log_settings):
        super(Mcpd, self).__init__(daemon_log_settings)
        self._meta_data['required_load_parameters'] = set()
        self._meta_data['required_json_kind'] =\
            'tm:sys:daemon-log-settings:mcpd:mcpdstate'


class Tmm(UnnamedResource):
    """BIG-IP® system daemon log settings clusterd resource

    The object only supports load and update because it is an
    unnamed resource.

    .. note::

        This is an unnamed resource so it has not ~Partition~Name pattern
        at the end of its URI.
    """
    def __init__(self, daemon_log_settings):
        super(Tmm, self).__init__(daemon_log_settings)
        self._meta_data['required_load_parameters'] = set()
        self._meta_data['required_json_kind'] =\
            'tm:sys:daemon-log-settings:tmm:tmmstate'



# coding=utf-8
#
# Copyright 2017 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""BIG-IP® system icall module

REST URI
    ``http://localhost/mgmt/tm/sys/icall``

GUI Path
    N/A

REST Kind
    ``tm:sys:icall:*``
"""

from f5.bigip.resource import Collection
from f5.bigip.resource import OrganizingCollection
from f5.bigip.resource import Resource


class Icall(OrganizingCollection):
    def __init__(self, sys):
        super(Icall, self).__init__(sys)
        self._meta_data['allowed_lazy_attributes'] = [
            Handler,
            Istats_Triggers,
            Scripts]


class Handler(OrganizingCollection):
    def __init__(self, Icall):
        super(Handler, self).__init__(Icall)
        self._meta_data['allowed_lazy_attributes'] = [
            Periodics,
            Perpetuals,
            Triggered_s]


class Periodics(Collection):
    def __init__(self, Handler):
        super(Periodics, self).__init__(Handler)
        self._meta_data['allowed_lazy_attributes'] = [Periodic]
        self._meta_data['attribute_registry'] =\
            {'tm:sys:icall:handler:periodic:periodicstate': Periodic}


class Periodic(Resource):
    def __init__(self, Periodics):
        super(Periodic, self).__init__(Periodics)
        self._meta_data['required_json_kind'] = \
            'tm:sys:icall:handler:periodic:periodicstate'
        self._meta_data['required_creation_parameters'].update(
            ('name', 'script', 'interval'))


class Perpetuals(Collection):
    def __init__(self, Handler):
        super(Perpetuals, self).__init__(Handler)
        self._meta_data['allowed_lazy_attributes'] = [Perpetual]
        self._meta_data['attribute_registry'] =\
            {'tm:sys:icall:handler:perpetual:perpetualstate': Perpetual}


class Perpetual(Resource):
    def __init__(self, Perpetuals):
        super(Perpetual, self).__init__(Perpetuals)
        self._meta_data['required_json_kind'] = \
            'tm:sys:icall:handler:perpetual:perpetualstate'
        self._meta_data['required_creation_parameters'].update(
            ('name', 'script'))


class Triggered_s(Collection):
    def __init__(self, Handler):
        super(Triggered_s, self).__init__(Handler)
        self._meta_data['allowed_lazy_attributes'] = [Triggered]
        self._meta_data['attribute_registry'] =\
            {'tm:sys:icall:handler:triggered:triggeredstate': Triggered}


class Triggered(Resource):
    def __init__(self, Triggered_s):
        super(Triggered, self).__init__(Triggered_s)
        self._meta_data['required_json_kind'] = \
            'tm:sys:icall:handler:triggered:triggeredstate'
        self._meta_data['required_creation_parameters'].update(
            ('name', 'script'))


class Scripts(Collection):
    def __init__(self, Icall):
        super(Scripts, self).__init__(Icall)
        self._meta_data['allowed_lazy_attributes'] = [Script]
        self._meta_data['attribute_registry'] = \
            {'tm:sys:icall:script:scriptstate': Script}


class Script(Resource):
    def __init__(self, Scripts):
        super(Script, self).__init__(Scripts)
        self._meta_data['required_json_kind'] = \
            'tm:sys:icall:script:scriptstate'
        self._meta_data['required_creation_parameters'].update(
            ('name',))


class Istats_Triggers(Collection):
    def __init__(self, Icall):
        super(Istats_Triggers, self).__init__(Icall)
        self._meta_data['allowed_lazy_attributes'] = [Istats_Trigger]
        self._meta_data['attribute_registry'] = \
            {'tm:sys:icall:istats-trigger:istats-triggerstate': Istats_Trigger}


class Istats_Trigger(Resource):
    def __init__(self, Istats_Triggers):
        super(Istats_Trigger, self).__init__(Istats_Triggers)
        self._meta_data['required_json_kind'] = \
            'tm:sys:icall:istats-trigger:istats-triggerstate'
        self._meta_data['required_creation_parameters'].update(
            ('name', 'eventName', 'istatsKey'))



# coding=utf-8
#
# Copyright 2014-2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""BIG-IP® Network vlan module.

REST URI
    ``http://localhost/mgmt/tm/net/vlan``

GUI Path
    ``Network --> Vlans``

REST Kind
    ``tm:net:vlan:*``
"""

from f5.bigip.mixins import ExclusiveAttributesMixin
from f5.bigip.resource import Collection
from f5.bigip.resource import Resource
from f5.sdk_exception import MissingUpdateParameter
from f5.sdk_exception import TagModeDisallowedForTMOSVersion

from distutils.version import LooseVersion


class Vlans(Collection):
    """BIG-IP® network Vlan collection."""
    def __init__(self, net):
        super(Vlans, self).__init__(net)
        self._meta_data['allowed_lazy_attributes'] = [Vlan]
        self._meta_data['attribute_registry'] =\
            {'tm:net:vlan:vlanstate': Vlan}


class Vlan(Resource):
    """BIG-IP® network Vlan resource."""
    def __init__(self, vlan_s):
        super(Vlan, self).__init__(vlan_s)
        self._meta_data['required_json_kind'] = 'tm:net:vlan:vlanstate'
        self._meta_data['attribute_registry'] =\
            {'tm:net:vlan:interfacescollectionstate': Interfaces_s}


class Interfaces_s(Collection):
    '''BIG-IP® network Vlan interface collection.

    .. note::
        Not to be confused with ``tm/mgmt/net/interface``.  This is object
        is actually called ``interfaces`` with an ``s`` by the BIG-IP's REST
        API.
    '''
    def __init__(self, vlan):
        super(Interfaces_s, self).__init__(vlan)
        self._meta_data['allowed_lazy_attributes'] = [Interfaces]
        self._meta_data['attribute_registry'] =\
            {'tm:net:vlan:interfaces:interfacesstate': Interfaces}


class Interfaces(Resource, ExclusiveAttributesMixin):
    """BIG-IP® network Vlan interface resource."""
    def __init__(self, interfaces_s):
        super(Interfaces, self).__init__(interfaces_s)
        # Vlan intefaces objects do not have a partition
        self._meta_data['required_json_kind'] =\
            'tm:net:vlan:interfaces:interfacesstate'
        # You cannot send both tagged and untagged attributes on update
        self._meta_data['exclusive_attributes'].append(('tagged', 'untagged'))

    def create(self, **kwargs):
        """Create the resource on the BIG-IP®.

        Uses HTTP POST to the `collection` URI to create a resource associated
        with a new unique URI on the device.

        As tagMode parameter will be required
        only if tagged is set to 'True'
        we have to use conditional to capture this logic during create.

        """

        self._check_tagmode_and_tmos_version(**kwargs)
        return self._create(**kwargs)

    def _check_tagmode_and_tmos_version(self, **kwargs):
        '''Raise an exception if tagMode in kwargs and tmos version < 11.6.0

        :param kwargs: dict -- keyword arguments for request
        :raises: TagModeDisallowedForTMOSVersion
        '''

        tmos_version = self._meta_data['bigip']._meta_data['tmos_version']
        if LooseVersion(tmos_version) < LooseVersion('11.6.0'):
            msg = "The parameter, 'tagMode', is not allowed against the " \
                "following version of TMOS: %s" % (tmos_version)
            if 'tagMode' in kwargs or hasattr(self, 'tagMode'):
                raise TagModeDisallowedForTMOSVersion(msg)

    def update(self, **kwargs):
        self._check_tagmode_and_tmos_version(**kwargs)
        if LooseVersion(self._meta_data['bigip']._meta_data['tmos_version']) \
                >= LooseVersion('11.6.0'):
            if 'tagged' in kwargs:
                if kwargs['tagged'] is True and 'tagMode' not in kwargs:
                    error = 'Missing tagMode parameter value.'
                    raise MissingUpdateParameter(error)
            if hasattr(self, 'tagged'):
                if getattr(self, 'tagged') is True and \
                   getattr(self, 'tagMode') == 'none':
                    error = 'Missing tagMode parameter value.'
                    raise MissingUpdateParameter(error)

        self._update(**kwargs)

        return self



# coding=utf-8
#
# Copyright 2015-2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import time

from f5.bigip.resource import AsmResource
from f5.bigip.resource import Collection
from icontrol.exceptions import iControlUnexpectedHTTPError


class Signatures_s(Collection):
    """BIG-IP® ASM Signatures collection."""
    def __init__(self, asm):
        super(Signatures_s, self).__init__(asm)
        self._meta_data['object_has_stats'] = False
        self._meta_data['allowed_lazy_attributes'] = [Signature]
        self._meta_data['attribute_registry'] = {
            'tm:asm:signatures:signaturestate': Signature
        }


class Signature(AsmResource):
    """BIG-IP® ASM Signature resource.

    note:: Only user created signatures can be modified/deleted.
           Default signatures are READ-ONLY
    """
    def __init__(self, signatures_s):
        super(Signature, self).__init__(signatures_s)
        self._meta_data['required_json_kind'] = 'tm:asm:signatures:signaturestate'
        self._meta_data['required_creation_parameters'].update(
            ('attackTypeReference', 'rule')
        )

    def create(self, **kwargs):
        """Custom creation logic to handle edge cases

        This shouldn't be needed, but ASM has a tendency to raise various errors that
        are painful to handle from a customer point-of-view. These errors are especially
        pronounced when doing things concurrently with asm.

        The error itself are described in their exception handler

        To address these failure, we try a number of exception handling cases to catch
        and reliably deal with the error.

        :param kwargs:
        :return:
        """
        ex = iControlUnexpectedHTTPError(
            "Failed to delete the signature"
        )
        for _ in range(0, 30):
            try:
                return self._create(**kwargs)
            except iControlUnexpectedHTTPError as ex:
                if self._check_exception(ex):
                    continue
                else:
                    raise
        raise ex

    def delete(self, **kwargs):
        """Custom deletion logic to handle edge cases

        This shouldn't be needed, but ASM has a tendency to raise various errors that
        are painful to handle from a customer point-of-view. These errors are especially
        pronounced when doing things concurrently with asm.

        The error itself are described in their exception handler

        To address these failure, we try a number of exception handling cases to catch
        and reliably deal with the error.

        :param kwargs:
        :return:
        """
        ex = iControlUnexpectedHTTPError(
            "Failed to delete the signature"
        )
        for _ in range(0, 30):
            try:
                return self._delete(**kwargs)
            except iControlUnexpectedHTTPError as ex:
                if self._check_exception(ex):
                    continue
                else:
                    raise
        raise ex

    def modify(self, **kwargs):
        ex = iControlUnexpectedHTTPError(
            "Failed to modify the signature"
        )
        for _ in range(0, 30):
            try:
                return self._modify(**kwargs)
            except iControlUnexpectedHTTPError as ex:
                if self._check_exception(ex):
                    continue
                else:
                    raise
        raise ex

    def update(self, **kwargs):
        ex = iControlUnexpectedHTTPError(
            "Failed to delete the signature"
        )
        for _ in range(0, 30):
            try:
                return self._update(**kwargs)
            except iControlUnexpectedHTTPError as ex:
                if self._check_exception(ex):
                    continue
                else:
                    raise
        raise ex

    def _check_exception(self, ex):
        """Check for exceptions in action responses

        In versions of ASM < v12, the REST API is quite unstable and therefore
        needs some additional supporting retries to ensure that actions function
        as expected. In particular versions 11.5.4 and 11.6.0 are affected.

        This method handles checking for various exceptions and allowing the
        given command to retry itself.

        :param ex:
        :return:
        """
        retryable = [
            # iControlUnexpectedHTTPError: 500 Unexpected Error: Internal Server Error ...
            # {
            #   "code": 500,
            #   "message": "Could not add_signature the Attack Signature.  "
            #              "Failed on insert to PLC.NEGSIG_SET_SIGNATURES "
            #              "(DBD::mysql::db do failed: Lock wait timeout exceeded; "
            #              "try restarting transaction)
            #
            'Lock wait timeout exceeded',

            # {
            #   "code": 500,
            #   "message": "DBD::mysql::db do failed: Deadlock found when "
            #              "trying to get lock; try restarting transaction"
            #
            'Deadlock found when',

            # {
            #   "code": 404,
            #   "message": "Could not add_signature the Attack Signature, "
            #              "internal data inconsistency was detected.",
            'internal data inconsistency',
        ]
        if any(x in str(ex) for x in retryable):
            time.sleep(3)
            return True
        elif 'errorStack' in ex:
            stack = ' '.join(ex['errorStack'])
            if any(x in stack for x in retryable):
                time.sleep(3)
                return True
            else:
                return False
        else:
            return False



# coding=utf-8
#
# Copyright 2015-2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os

from f5.bigip.mixins import AsmFileMixin
from f5.bigip.resource import OrganizingCollection
from f5.bigip.resource import PathElement
from f5.sdk_exception import FileMustNotHaveDotISOExtension


class File_Transfer(OrganizingCollection):
    """BIG-IP® ASM File Transfer collection."""
    def __init__(self, tm):
        super(File_Transfer, self).__init__(tm)
        self._meta_data['object_has_stats'] = False
        self._meta_data['minimum_version'] = '11.6.0'
        self._meta_data['allowed_lazy_attributes'] = [
            Uploads,
            Downloads,
        ]


class Uploads(PathElement, AsmFileMixin):
    """A file upload resource."""
    def __init__(self, file_transfer):
        super(Uploads, self).__init__(file_transfer)
        self._meta_data['object_has_stats'] = False
        self.file_bound_uri = ''

    def upload_file(self, filepathname, **kwargs):
        filename = os.path.basename(filepathname)
        if os.path.splitext(filename)[-1] == '.iso':
            raise FileMustNotHaveDotISOExtension(filename)
        self.file_bound_uri = self._meta_data['uri'] + filename
        self._upload_file(filepathname, **kwargs)


class Downloads(PathElement, AsmFileMixin):
    """A file download resource."""
    def __init__(self, file_transfer):
        super(Downloads, self).__init__(file_transfer)
        self._meta_data['object_has_stats'] = False
        self.file_bound_uri = ''

    def download_file(self, filepathname):
        filename = os.path.basename(filepathname)
        self.file_bound_uri = self._meta_data['uri'] + filename
        self._download_file(filepathname)



#!/usr/bin/env python
# Copyright 2017 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
*** This script, install_test.py, is intentionally titled in this way so that
it does not get auto-grabbed by the tox application. ***

This script is both a standalone as well as a callable/integratable module;
thus, for terms of use, it is a modulino.  It is used to test both .deb and
.rpm build packages.

Future coverage is a always a problem that demands continuous maintenance.  As
such, this script looks to, when run directly, the build.py's
expected_operations-stored variable.

When imported, it is suggested to use the install_test class as an API.
"""

import errno
import glob
import os
import re
import sys
import traceback

from collections import deque
from collections import namedtuple
from inspect import currentframe as cf
from inspect import getframeinfo as gfi

from . import build_expectations

from . terminal import terminal
from . build_exceptions import TestError

# Globals:
builds = build_expectations.Builds()
OperationSettings = builds.OperationSettings
expected_operations = builds.expected_operations
Operation = namedtuple('Operation', 'os_type, os, version, pkg, dockerfl')


class InstallTest(object):
    """InstallTest

This is a package test API.  Basically, this class can be used to test both
rpm and deb build_pkgs packages.  These packages are tested natively, and
there must already exist a:

./*-dist/Docker/<os>/install_test/Dockerfile*<version>

For supportaiblity and expandability, 'trusty' can be interpretted as 14.04
for ubuntu; thus, this script does attempt to swap the word 'trusty' for any
given version when searching for the docker file.  The term 'trusty' can be
used interchangably; however, the actual version number will be searched for
first.
    """
    __os_types = []  # we never use this, but the caller can extract it

    def __init__(self, os_types=[], expected_opts=tuple([])):
        """Creation

install_test(os_types=[], expected_opts=tuple([]), )
    This will construct the object.

    Options:
        os_types - must be a list of operating system types that already exist
            within the expected_operations tuple([]), but are
            specifically tested against **instead of** all of them
        expected_opts - overwrites the tuple([]) expected_opterations for this
            object's instance; thus, replaces the default expected_operations
            that comes from the build_pkgs.py script
    Options are not suggested for general practice of automation, but they can
    be useful for testing purposes.
        """
        self._set_failure_reason = None
        self._set_expected_opts = expected_opts
        self._set_os_types = os_types
        self.collect_dist_dir()
        self.collect_pkgs()

    @property
    def expected_opts(self):
        """expected_opts

Defines the expected_opts tuple([OperationSettings])

This object attr is immutable.
        """
        return self.__expected_opts

    @property
    def failure_reason(self):
        """failure_reason

This attribute informs the caller that a test failure occurred.
        """
        return self.__failure_reason

    @property
    def os_types(self):
        """os_types

Defines the needed os_types to be tested that will be tested during runtime.
Format for this object should follow tuple([str, str,...]).

This object attr is immutable.
        """
        return self.__os_types

    @expected_opts.setter
    def _set_expected_opts(self, opts):
        err_msg = "expected_opts option contains a structure not recognized"
        self.__expected_opts = expected_operations
        if opts:
            if not isinstance(opts, tuple) or \
                    not isinstance(opts[0][0], OperationSettings):
                raise TestError(msg=err_msg, frame=gfi(cf()),
                                errnum=errno.ESPIPE)
            self.__expected_opts = opts

    @failure_reason.setter
    def _set_failure_reason(self, failure_reason):
        self.__failure_reason = failure_reason

    @os_types.setter
    def _set_os_types(self, types):
        if not isinstance(types, list):
            raise TypeError("os_types is not a list!")
        elif not types:
            return
        opts = list()
        for opt in self.expected_opts:
            if opt.os_type in types:
                opts.append(opt)
            else:
                print("Excluding: %s from testing" % opt.os_type)
        self._set_expected_opts = tuple(opts)
        self.__os_types = types

    def collect_dist_dir(self):
        """collect_dist_dir

Collects the distribution directory from where this script was located.
        """
        working = os.path.dirname(os.path.dirname(sys.argv[0]))
        working = no_ddots(working)
        working = os.path.abspath('..') if not working else working
        self.working = working

    def collect_pkgs(self):
        """collect_pkgs

Collects the packages found based upon the discovered dist dir and the list of
tests that are lined up.
        """
        pkgs = deque()
        dist = self.working
        docker = self.working + "/Docker/"
        for opt in self.expected_opts:
            install_test = docker + "/%s/install_test" % opt.os_type
            if opt.os_category == 'debian':
                # not setting a default b/c
                pkg_search = dist + "/deb_dist/*.deb"
                pkg_re = re.compile('_(\d+)_all\.deb')
            elif opt.os_category == 'redhat':
                pkg_search = dist + "/rpms/build/*.rpm"
                pkg_re = re.compile('el(\d+)\.noarch')
            else:
                raise TestError(opt.os_type, frame=gfi(cf()),
                                errnum=errno.ESPIPE,
                                msg='opt.os_type(%s) is not recognized!')
            entropy = glob.glob(pkg_search)
            if entropy:
                for version in opt.versions:
                    found = False
                    for pkg in entropy:
                        basename = os.path.basename(pkg)
                        match = pkg_re.search(basename)
                        if match:
                            found = match.group(1)
                            if found == version.replace('.', ''):
                                dockerfl = \
                                    install_test + "/Dockerfile.*%s" % version
                                entropy = glob.glob(dockerfl)
                                dockerfl = entropy[0] if entropy else ''
                                trusty = install_test + "/Dockerfile.trusty"
                                dockerfl = dockerfl \
                                    if os.path.isfile(dockerfl) else trusty
                                pkgs.append(Operation(opt.os_category,
                                                      opt.os_type, version,
                                                      pkg, dockerfl))
                                found = True
                                break
                        break
                    if not found:
                        self._set_failure_reason = \
                            TestError(opt.os_type, opt.version,
                                      frame=gfi(cf()), errno=errno.ESPIPE,
                                      msg='No pkg found built for')
                        print(str(self.failure_reason))
        self.tst_pkgs = pkgs

    def execute_tests(self):
        """execute_tests

Performs the series of tests specified at object creation via either:
    - Default
        expected_operations
    - Specified os_types
        build_pkgs.expected_oeprations[*].os_type
    - Specific expected_opts
        Custom tuple([OperationSettings])
        """
        if self.failure_reason:
            raise self.failure_reason
        print("Testing Packages...")
        for pkg in self.tst_pkgs:
            self.test(pkg)

    def test(self, pkg):
        """test

Performs a test against a particular install_test.Operation namedtuple.

Outside:
    obj.test(Operation(os_category, os_type, version, pkg_name, dockerfile))

    Currently, os_category is not used and can be a blank str.
        """
        container = "%s%s-%s-pkg-tester" % \
            (pkg.os, pkg.version, os.path.basename(pkg.pkg))
        build_cmd = "docker build -t %s -f %s %s" % \
            (container, pkg.dockerfl, os.path.dirname(pkg.dockerfl))
        wkg = os.path.dirname(os.path.abspath(self.working))
        tst_cmd = "docker run --privileged --rm -v %s:/var/wdir %s %s" % \
            (wkg, container,
             re.sub('.+/([^\/]+-dist/.+)', '/var/wdir/\g<1>', pkg.pkg))
        result = terminal([build_cmd], multi_stmt=True)
        frame = None
        msg = ''
        if result.succeeded:
            result = terminal([tst_cmd], multi_stmt=True)
            if result.succeeded:
                print("Package: %s Test succeeded!" % pkg.pkg)
            else:
                print(result.cli)
                frame = gfi(cf())
                msg = "Failed to perform install test"
        else:
            print(result.cli)
            frame = gfi(cf())
            msg = "Failed to build docker container to test with"
        if frame:
            self._set_failure_reason = \
                TestError(pkg.pkg, msg=msg, frame=frame, errnum=errno.ESPIPE)
            raise self.failure_reason


def no_ddots(wkg):
    """no_ddots

Returns a directory path that does not contain any regression movments in it.
That is to say the /../ FS object 'previous directory' pointer.
    """
    lower = deque()
    if '/../' in wkg:
        triage = wkg.split('/')
        for item in triage:
            if '..' == item:
                lower.pop()
            else:
                lower.append(item)
        working = '/'
        while lower:
            working = working + "%s/" + lower.popleft()
    else:
        working = wkg
    return working if os.path.isdir(working) else wkg


def main():
    """main

The standalone entrypoint to the script.  This will initiate build pkg's tests
on its own.
    """
    try:
        tests = InstallTest()
        tests.execute_tests()
    except TestError as Error:
        print(str(Error))
        exit(1)
        traceback.print_exc()
    except Exception as Error:
        print(str(Error))
        TestError(Error, frame=gfi(cf()), errno=-1)
        print(str(Error))
        traceback.print_exc()


if __name__ == '__main__':
    main()



# coding=utf-8
#
# Copyright 2015-2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""BIG-IP® auth module

REST URI
    ``http://localhost/mgmt/tm/auth/``

GUI Path
    ``System --> Users``

REST Kind
    ``tm:auth:*``
"""


from f5.bigip.resource import OrganizingCollection
from f5.bigip.tm.auth.cert_ldap import Cert_Ldaps
from f5.bigip.tm.auth.ldap import Ldaps
from f5.bigip.tm.auth.partition import Partitions
from f5.bigip.tm.auth.password_policy import Password_Policy
from f5.bigip.tm.auth.radius import Radius_s
from f5.bigip.tm.auth.radius_server import Radius_Servers
from f5.bigip.tm.auth.remote_role import Remote_Role
from f5.bigip.tm.auth.remote_user import Remote_User
from f5.bigip.tm.auth.source import Source
from f5.bigip.tm.auth.tacacs import Tacacs_s
from f5.bigip.tm.auth.user import Users


class Auth(OrganizingCollection):
    def __init__(self, tm):
        super(Auth, self).__init__(tm)
        self._meta_data['allowed_lazy_attributes'] = [
            Cert_Ldaps,
            Ldaps,
            Partitions,
            Password_Policy,
            Radius_s,
            Radius_Servers,
            Remote_Role,
            Remote_User,
            Source,
            Tacacs_s,
            Users
        ]



# coding=utf-8
#
# Copyright 2015-2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from f5.bigip.resource import Collection as BigIpCollection
from f5.bigip.resource import HTTPError
from f5.bigip.resource import OrganizingCollection as BigIpOrganizingCollection
from f5.bigip.resource import PathElement as BigIpPathElement
from f5.bigip.resource import Resource as BigIpResource
from f5.bigip.resource import ResourceBase as BigIpResourceBase
from f5.bigip.resource import UnnamedResource as BigIpUnnamedResource
from f5.sdk_exception import UnsupportedOperation
from f5.sdk_exception import URICreationCollision


class PathElement(BigIpPathElement):
    """Base class to represent a URI path element that does not contain data.

    The BIG-IP® iControl REST API has URIs that are made up of path components
    that do not return data when they are queried.  This class represents
    those elements and does not support any of the CURDLE methods that
    the other objects do.
    """

    def __init__(self, container):
        super(PathElement, self).__init__(container)
        self._meta_data['minimum_version'] = '5.3.0'


class Resource(BigIpResource, PathElement):
    def __init__(self, container):
        super(Resource, self).__init__(container)
        self._meta_data['required_load_parameters'] = {'id', }


class ResourceBase(BigIpResourceBase, PathElement):
    pass


class OrganizingCollection(BigIpOrganizingCollection, ResourceBase):
    pass


class UnnamedResource(BigIpUnnamedResource, ResourceBase):
    pass


class Collection(BigIpCollection, ResourceBase):
    pass


class TaskResource(Resource):
    def __init__(self, container):
        """Call to create a client side object to represent a service URI.

        Call _create or _load for a Resource to have a self._meta_data['uri']!
        """
        super(TaskResource, self).__init__(container)
        # Asm endpoints require object 'id' which is a hash created by BIGIP
        # when object is created.
        self._meta_data['required_load_parameters'] = {'id', }
        # No ASM endpoint supports Stats
        self._meta_data['object_has_stats'] = False

    def _load(self, **kwargs):
        """wrapped with load, override that in a subclass to customize"""
        if 'uri' in self._meta_data:
            error = "There was an attempt to assign a new uri to this " \
                    "resource, the _meta_data['uri'] is %s and it should" \
                    " not be changed." % (self._meta_data['uri'])
            raise URICreationCollision(error)
        requests_params = self._handle_requests_params(kwargs)
        self._check_load_parameters(**kwargs)
        kwargs['uri_as_parts'] = True
        refresh_session = self._meta_data['bigip']._meta_data['icr_session']
        uri = self._meta_data['container']._meta_data['uri']
        endpoint = kwargs.pop('id', '')
        # Popping name kwarg as it will cause the uri to be invalid. We only
        # require id parameter
        kwargs.pop('name', '')
        base_uri = uri + endpoint + '/'
        kwargs.update(requests_params)
        for key1, key2 in self._meta_data['reduction_forcing_pairs']:
            kwargs = self._reduce_boolean_pair(kwargs, key1, key2)
        response = refresh_session.get(base_uri, **kwargs)
        # Make new instance of self
        return self._produce_instance(response)

    def load(self, **kwargs):
        """Load an already configured service into this instance.

        This method uses HTTP GET to obtain a resource from the BIG-IP®.

        ..
            The URI of the target service is constructed from the instance's
            container and **kwargs.
            kwargs typically for ASM requires "id" in majority of cases,
            as object links in ASM are using hash(id) instead of names,
            this may, or may not, be true for a specific service.

        :param kwargs: typically contains "id"
        NOTE: If kwargs has a 'requests_params' key the corresponding dict will
        be passed to the underlying requests.session.get method where it will
        be handled according to that API. THIS IS HOW TO PASS QUERY-ARGS!
        :returns: a Resource Instance (with a populated _meta_data['uri'])
        """
        return self._load(**kwargs)

    def _delete(self, **kwargs):
        """wrapped with delete, override that in a subclass to customize """
        requests_params = self._handle_requests_params(kwargs)

        delete_uri = self._meta_data['uri']
        session = self._meta_data['bigip']._meta_data['icr_session']
        response = session.delete(delete_uri, **requests_params)
        if response.status_code == 200 or 201:
            self.__dict__ = {'deleted': True}

    def delete(self, **kwargs):
        """Delete the Task resource on the BIG-IP®.

        Uses HTTP DELETE to delete the Task resource on the BIG-IP®.

        After this method is called, and status_code 200 or 201 response is
        received ``instance.__dict__`` is replace with ``{'deleted': True}``

        :param kwargs: The only current use is to pass kwargs to the requests
        API. If kwargs has a 'requests_params' key the corresponding dict will
        be passed to the underlying requests.session.delete method where it
        will be handled according to that API. THIS IS HOW TO PASS QUERY-ARGS!
        """
        # Need to implement checking for ? here.
        self._delete(**kwargs)
        # Need to implement correct teardown here.

    def exists(self, **kwargs):
        """Check for the existence of the Task object on the BIG-IP

        Sends an HTTP GET to the URI of the ASM object and if it fails with
        a :exc:~requests.HTTPError` exception it checks the exception for
        status code of 404 and returns :obj:`False` in that case.

        If the GET is successful it returns :obj:`True`.

        For any other errors are raised as-is.

        :param kwargs: Keyword arguments required to get objects
        NOTE: If kwargs has a 'requests_params' key the corresponding dict will
        be passed to the underlying requests.session.get method where it will
        be handled according to that API. THIS IS HOW TO PASS QUERY-ARGS!
        :returns: bool -- The objects exists on BIG-IP® or not.
        :raises: :exc:`requests.HTTPError`, Any HTTP error that was not status
                 code 404.
        """
        requests_params = self._handle_requests_params(kwargs)
        self._check_load_parameters(**kwargs)
        kwargs['uri_as_parts'] = True
        session = self._meta_data['bigip']._meta_data['icr_session']
        uri = self._meta_data['container']._meta_data['uri']
        endpoint = kwargs.pop('id', '')
        # Popping name kwarg as it will cause the uri to be invalid
        kwargs.pop('name', '')
        base_uri = uri + endpoint + '/'
        kwargs.update(requests_params)
        try:
            session.get(base_uri, **kwargs)
        except HTTPError as err:
            if err.response.status_code == 404:
                return False
            else:
                raise
        return True

    def update(self, **kwargs):
        """Update is not supported for ASM Resources

                :raises: UnsupportedOperation
        """
        raise UnsupportedOperation(
            "%s does not support the update method" % self.__class__.__name__
        )



# coding=utf-8
#
#  Copyright 2017 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


from f5.bigip.resource import Collection
from f5.bigip.resource import OrganizingCollection
from f5.bigip.resource import Resource
from f5.sdk_exception import UnsupportedMethod
from f5.sdk_exception import UnsupportedOperation
from f5.sdk_exception import URICreationCollision


class Authn(OrganizingCollection):
    def __init__(self, shared):
        super(Authn, self).__init__(shared)
        self._meta_data['allowed_lazy_attributes'] = [
            Roots
        ]


class Roots(Collection):
    def __init__(self, authn):
        super(Roots, self).__init__(authn)
        self._meta_data['allowed_lazy_attributes'] = [Root]

    def get_collection(self, **kwargs):
        raise UnsupportedMethod(
            "%s does not support get_collection" % self.__class__.__name__
        )


class Root(Resource):
    def __init__(self, roots):
        super(Root, self).__init__(roots)
        self._meta_data['required_json_kind'] = 'shared:authn:authrootitemstate'

        # The required parameters are a little vague. It turns out that the "user"
        # value that is required is the "link" to the ID
        self._meta_data['required_creation_parameters'] = {'oldPassword', 'newPassword'}

    def _create(self, **kwargs):
        """wrapped by `create` override that in subclasses to customize"""
        if 'uri' in self._meta_data:
            error = "There was an attempt to assign a new uri to this "\
                    "resource, the _meta_data['uri'] is %s and it should"\
                    " not be changed." % (self._meta_data['uri'])
            raise URICreationCollision(error)
        self._check_exclusive_parameters(**kwargs)
        requests_params = self._handle_requests_params(kwargs)
        self._minimum_one_is_missing(**kwargs)
        self._check_create_parameters(**kwargs)
        kwargs = self._check_for_python_keywords(kwargs)

        # Reduce boolean pairs as specified by the meta_data entry below
        for key1, key2 in self._meta_data['reduction_forcing_pairs']:
            kwargs = self._reduce_boolean_pair(kwargs, key1, key2)

        # Make convenience variable with short names for this method.
        _create_uri = self._meta_data['container']._meta_data['uri']
        session = self._meta_data['bigip']._meta_data['icr_session']

        kwargs = self._prepare_request_json(kwargs)

        # Invoke the REST operation on the device.
        response = session.post(_create_uri, json=kwargs, **requests_params)

        # Make new instance of self
        result = self._produce_instance(response)
        return result

    def _local_update(self, rdict):
        super(Root, self)._local_update(rdict)

        # This API returns no kind, so we need to make our own
        self.__dict__.update(dict(kind='shared:authn:authrootitemstate'))

        # This API returns no selfLink, so we need to make our own
        tmos_version = self._meta_data['bigip']._meta_data['tmos_version']
        self_link = 'https://localhost/mgmt/shared/authn/root?ver={0}'.format(tmos_version)
        self.__dict__.update(dict(selfLink=self_link))

    def update(self, **kwargs):
        raise UnsupportedOperation(
            "%s does not support update" % self.__class__.__name__
        )

    def load(self, **kwargs):
        raise UnsupportedOperation(
            "%s does not support load" % self.__class__.__name__
        )

    def modify(self, **kwargs):
        raise UnsupportedOperation(
            "%s does not support modify" % self.__class__.__name__
        )

    def delete(self, **kwargs):
        raise UnsupportedOperation(
            "%s does not support delete" % self.__class__.__name__
        )



# coding=utf-8
#
# Copyright 2017 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from distutils.version import LooseVersion
from f5.bigip.resource import AsmResource
from f5.bigip.resource import Collection


class UrlParametersCollection(Collection):
    """BIG-IP® ASM Urls Parameters sub-collection."""
    def __init__(self, urls_s):
        self.__class__.__name__ = 'Parameters_s'
        super(UrlParametersCollection, self).__init__(urls_s)
        self._meta_data['object_has_stats'] = False
        self._meta_data['allowed_lazy_attributes'] = [Parameter]
        self._meta_data['required_json_kind'] = 'tm:asm:policies:urls:parameters:parametercollectionstate'
        self._meta_data['attribute_registry'] = {
            'tm:asm:policies:urls:parameters:parameterstate': Parameter
        }


class ParametersCollection(Collection):
    """BIG-IP® ASM Policies Parameters sub-collection."""
    def __init__(self, policy):
        self.__class__.__name__ = 'Parameters_s'
        super(ParametersCollection, self).__init__(policy)
        self._meta_data['object_has_stats'] = False
        self._meta_data['allowed_lazy_attributes'] = [Parameter]
        self._meta_data['required_json_kind'] = 'tm:asm:policies:parameters:parametercollectionstate'
        self._meta_data['attribute_registry'] = {
            'tm:asm:policies:parameters:parameterstate': Parameter
        }


class Parameters_s(object):
    """As Parameters classes are used twice as a sub-collection.


    We need to utilize __new__ method in order to keep the user
    interface consistent.

    """
    def __new__(cls, container):
        from f5.bigip.tm.asm.policies import Policy
        from f5.bigip.tm.asm.policies.urls import Url
        if isinstance(container, Policy):
            return ParametersCollection(container)
        if isinstance(container, Url):
            return UrlParametersCollection(container)


class Parameter(object):
    """As Parameter classes are used twice as a sub-collection.


    We need to utilize __new__ method in order to keep the user
    interface consistent.
    """

    def __new__(cls, container):
        if isinstance(container, ParametersCollection):
            return ParametersResource(container)
        if isinstance(container, UrlParametersCollection):
            return UrlParametersResource(container)


class UrlParametersResource(AsmResource):
    """BIG-IP® ASM Urls Parameters resource."""
    def __init__(self, urls_s):
        self.__class__.__name__ = 'Parameter'
        super(UrlParametersResource, self).__init__(urls_s)
        self.tmos_v = urls_s._meta_data['bigip']._meta_data['tmos_version']
        self._meta_data['required_json_kind'] = 'tm:asm:policies:urls:parameters:parameterstate'

    def create(self, **kwargs):
        """Custom create method for v12.x and above.


        Change of behavior in v12 where the returned selfLink is different
        from target resource, requires us to append URI after object is
        created. So any modify() calls will not lead to json kind
        inconsistency when changing the resource attribute.

        See issue #844
        """
        if LooseVersion(self.tmos_v) < LooseVersion('12.0.0'):
            return self._create(**kwargs)
        else:
            new_instance = self._create(**kwargs)
            tmp_name = str(new_instance.id)
            tmp_path = new_instance._meta_data['container']._meta_data['uri']
            finalurl = tmp_path + tmp_name
            new_instance._meta_data['uri'] = finalurl
            return new_instance


class ParametersResource(AsmResource):
    """BIG-IP® ASM Urls Parameters resource."""
    def __init__(self, policy):
        self.__class__.__name__ = 'Parameter'
        super(ParametersResource, self).__init__(policy)
        self._meta_data['required_json_kind'] = 'tm:asm:policies:parameters:parameterstate'



# coding=utf-8
#
# Copyright 2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""BIG-IP® system file module

REST URI
    ``http://localhost/mgmt/tm/sys/file``

GUI Path
    N/A

REST Kind
    ``tm:sys:file:*``
"""

from distutils.version import LooseVersion
from f5.bigip.resource import Collection
from f5.bigip.resource import OrganizingCollection
from f5.bigip.resource import Resource
from f5.sdk_exception import UnsupportedMethod


class File(OrganizingCollection):
    """BIG-IP® System sys file collection."""
    def __init__(self, sys):
        super(File, self).__init__(sys)
        self._meta_data['allowed_lazy_attributes'] = [
            Data_Groups,
            Ifiles,
            Ssl_Certs,
            Ssl_Csrs,
            Ssl_Crls,
            Ssl_Keys]


class Data_Groups(Collection):
    """BIG-IP® System sys file data-groups collection."""
    def __init__(self, File):
        super(Data_Groups, self).__init__(File)
        self._meta_data['allowed_lazy_attributes'] = [Data_Group]
        self._meta_data['attribute_registry'] =\
            {'tm:sys:file:data-group:data-groupstate': Data_Group}


class Data_Group(Resource):
    """BIG-IP® System sys file data-groups resource."""
    def __init__(self, data_groups):
        super(Data_Group, self).__init__(data_groups)
        self._meta_data['required_json_kind'] =\
            'tm:sys:file:data-group:data-groupstate'
        self._meta_data['required_creation_parameters'].update(
            ('name', 'sourcePath', 'type'))

    def update(self, **kwargs):
        if LooseVersion(self._meta_data['bigip']._meta_data['tmos_version']) \
                < LooseVersion('12.0.0'):
            if 'type' in self.__dict__:
                del self.__dict__['type']
        return self._update(**kwargs)


class Ifiles(Collection):
    """BIG-IP® System sys file iFiles collection."""
    def __init__(self, File):
        super(Ifiles, self).__init__(File)
        self._meta_data['allowed_lazy_attributes'] = [Ifile]
        self._meta_data['attribute_registry'] = \
            {'tm:sys:file:ifile:ifilestate': Ifile}


class Ifile(Resource):
    """BIG-IP® System sys file iFiles resource."""
    def __init__(self, ifiles):
        super(Ifile, self).__init__(ifiles)
        self._meta_data['required_json_kind'] =\
            'tm:sys:file:ifile:ifilestate'
        self._meta_data['required_creation_parameters'].update(
            ('name', 'sourcePath'))

    def modify(self, **kwargs):
        '''Modify is not supported for iFiles

        :raises: UnsupportedOperation
        '''
        raise UnsupportedMethod(
            "%s does not support the update method" % self.__class__.__name__
        )


class Ssl_Certs(Collection):
    """BIG-IP® System sys file ssl-certs collection."""
    def __init__(self, File):
        super(Ssl_Certs, self).__init__(File)
        self._meta_data['allowed_lazy_attributes'] = [Ssl_Cert]
        self._meta_data['attribute_registry'] =\
            {'tm:sys:file:ssl-cert:ssl-certstate': Ssl_Cert}


class Ssl_Cert(Resource):
    """BIG-IP® System sys file ssl-certs resource."""
    def __init__(self, ssl_certs):
        super(Ssl_Cert, self).__init__(ssl_certs)
        self._meta_data['required_json_kind'] =\
            'tm:sys:file:ssl-cert:ssl-certstate'
        self._meta_data['required_creation_parameters'].update(
            ('name', 'sourcePath'))

    def modify(self, **kwargs):
        '''Modify is not supported for iFiles

        :raises: UnsupportedOperation
        '''
        raise UnsupportedMethod(
            "%s does not support the update method" % self.__class__.__name__
        )


class Ssl_Crls(Collection):
    """BIG-IP® System sys file ssl-crls collection."""
    def __init__(self, File):
        super(Ssl_Crls, self).__init__(File)
        self._meta_data['allowed_lazy_attributes'] = [Ssl_Crl]
        self._meta_data['attribute_registry'] =\
            {'tm:sys:file:ssl-crl:ssl-crlstate': Ssl_Crl}


class Ssl_Crl(Resource):
    """BIG-IP® System sys file ssl-crls resource."""
    def __init__(self, ssl_crls):
        super(Ssl_Crl, self).__init__(ssl_crls)
        self._meta_data['required_json_kind'] =\
            'tm:sys:file:ssl-crl:ssl-crlstate'
        self._meta_data['required_creation_parameters'].update(
            ('name', 'sourcePath'))

    def modify(self, **kwargs):
        '''Modify is not supported for iFiles

        :raises: UnsupportedOperation
        '''
        raise UnsupportedMethod(
            "%s does not support the update method" % self.__class__.__name__
        )


class Ssl_Csrs(Collection):
    """BIG-IP® System sys file ssl-csrs collection."""
    def __init__(self, File):
        super(Ssl_Csrs, self).__init__(File)
        self._meta_data['allowed_lazy_attributes'] = [Ssl_Csr]
        self._meta_data['attribute_registry'] =\
            {'tm:sys:file:ssl-csr:ssl-csrstate': Ssl_Csr}
        self._meta_data['minimum_version'] = '12.0.0'


class Ssl_Csr(Resource):
    """BIG-IP® System sys file ssl-csrs resource."""
    def __init__(self, ssl_csrs):
        super(Ssl_Csr, self).__init__(ssl_csrs)
        self._meta_data['required_json_kind'] =\
            'tm:sys:file:ssl-csr:ssl-csrstate'
        self._meta_data['required_creation_parameters'].update(
            ('name', 'sourcePath'))

    def modify(self, **kwargs):
        '''Modify is not supported for iFiles

        :raises: UnsupportedOperation
        '''
        raise UnsupportedMethod(
            "%s does not support the update method" % self.__class__.__name__
        )


class Ssl_Keys(Collection):
    """BIG-IP® System sys file ssl-keys collection."""
    def __init__(self, File):
        super(Ssl_Keys, self).__init__(File)
        self._meta_data['allowed_lazy_attributes'] = [Ssl_Key]
        self._meta_data['attribute_registry'] =\
            {'tm:sys:file:ssl-key:ssl-keystate': Ssl_Key}


class Ssl_Key(Resource):
    """BIG-IP® System sys file ssl-keys resource."""
    def __init__(self, ssl_keys):
        super(Ssl_Key, self).__init__(ssl_keys)
        self._meta_data['required_json_kind'] =\
            'tm:sys:file:ssl-key:ssl-keystate'
        self._meta_data['required_creation_parameters'].update(
            ('name', 'sourcePath'))

    def modify(self, **kwargs):
        '''Modify is not supported for iFiles

        :raises: UnsupportedOperation
        '''
        raise UnsupportedMethod(
            "%s does not support the update method" % self.__class__.__name__
        )



from __future__ import absolute_import
# Copyright 2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from f5.devtools.code_generator import DEVICECONFDIR

import json
import os

from pprint import pprint as pp
from urlparse import urlparse


class OCCrawler(object):
    def __init__(self, bigip, OC_path_element):
        self.bigip = bigip
        self.session = self.bigip._meta_data[u"icr_session"]
        self.uri = self.bigip._meta_data['uri'] + OC_path_element
        self.configs = [self.session.get(self.uri).json()]
        self.build_referenced_uris()

    def _get_uri_from_OC_item(self, item):
        if u"reference" in item and u"link" in item[u"reference"]:
            return item[u"reference"][u"link"]\
                .replace("localhost",
                         self.bigip._meta_data[u"hostname"])

    def build_referenced_uris(self):
        self.referenced = []
        for item in self.configs[0][u"items"]:
            self.referenced.append(self._get_uri_from_OC_item(item))

    def get_referenced_configs(self):
        for uri in self.referenced:
            self.configs.append(self.session.get(uri).json())


class ConfigWriter(object):
    def __init__(self, config_list, complete_oc_name):
        self.oc_name = complete_oc_name
        self.oc_basename = self.oc_name.split('/')[-1]
        self.configs = config_list

    def _get_fname(self, conf):
        sl = conf[u"selfLink"]
        scheme, netloc, path, params, qargs, frags = urlparse(sl)
        ps = path.split('/')
        if ps[-1] == self.oc_basename:
            return self.oc_basename + '_GET'
        else:
            return self.oc_basename + '_' + ps[-1] + '_GET'

    def dump_configs(self):
        for conf in self.configs:
            fname = self._get_fname(conf)
            if not os.path.exists(os.path.join(DEVICECONFDIR, fname)):
                outname = os.path.join(DEVICECONFDIR, fname) + ".json"
                with open(outname, 'w') as fh:
                    json.dump(conf, fh)


def main():
    from f5.bigip import BigIP
    b = BigIP('10.190.5.7', 'admin', 'admin')
    occrawler = OCCrawler(b, 'ltm/persistence')
    pp(occrawler.referenced)
    occrawler.get_referenced_configs()
    pp(occrawler.configs)
    config_writer = ConfigWriter(occrawler.configs, u"ltm/persistence")
    config_writer.dump_configs()

if __name__ == '__main__':
    main()



# coding=utf-8
#
#  Copyright 2014-2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""BIG-IP® Local Traffic Manager (LTM) Nat module.

REST URI
    ``http://localhost/mgmt/tm/ltm/nat``

GUI Path
    ``Local Traffic --> Nat``

REST Kind
    ``tm:ltm:nat:*``
"""

from f5.bigip.mixins import ExclusiveAttributesMixin
from f5.bigip.resource import Collection
from f5.bigip.resource import Resource
from f5.sdk_exception import MissingRequiredCreationParameter


class Nats(Collection):
    """BIG-IP® LTM Nat collection object"""
    def __init__(self, ltm):
        super(Nats, self).__init__(ltm)
        self._meta_data['allowed_lazy_attributes'] = [Nat]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:nat:natstate': Nat}


class Nat(Resource, ExclusiveAttributesMixin):
    """BIG-IP® LTM Nat collection resource"""
    def __init__(self, nat_s):
        super(Nat, self).__init__(nat_s)
        self._meta_data['required_creation_parameters'].update(
            ('originatingAddress', 'translationAddress', 'partition'))
        self._meta_data['required_json_kind'] = 'tm:ltm:nat:natstate'
        self._meta_data['exclusive_attributes'].append(('enable', 'disable'))

    def create(self, **kwargs):
        """Create the resource on the BIG-IP®.

        Uses HTTP POST to the `collection` URI to create a resource associated
        with a new unique URI on the device.

        ..
            If you do a create with inheritedTrafficGroup set to 'false' you
            must also have a trafficGroup.  This pattern generalizes like so:
            If the presence of a param implies an additional required param,
            then simply
            self._meta_data['required_creation_params'].update(IMPLIED),
            before the call to self._create(**kwargs), wherein req params are
            checked.

            We refer to this property as "implied-required parameters" because
            the presence of one parameter, or parameter value (e.g.
            inheritedTrafficGroup), implies that another parameter is required.

        .. note::
            If you are creating with ``inheritedTrafficGroup` set to
            :obj:`False` you just also have a `trafficGroup`.

        :param kwargs: All the key-values needed to create the resource
        :returns: ``self`` - A python object that represents the object's
                  configuration and state on the BIG-IP®.

        """
        itg = kwargs.get('inheritedTrafficGroup', None)
        if itg and itg == 'false':
            self._meta_data['required_creation_parameters'].\
                update(('trafficGroup',))
            try:
                if not kwargs['trafficGroup']:
                    raise MissingRequiredCreationParameter(
                        "trafficGroup must not be falsey but it's: %r"
                        % kwargs['trafficGroup'])
            except KeyError:
                pass
        new_instance = self._create(**kwargs)
        return new_instance

    def update(self, **kwargs):
        # This is an example implementation of read-only params
        stash_translation_address = self.__dict__.pop('translationAddress')
        if 'enabled' in self.__dict__ and 'enabled' not in kwargs:
            kwargs['enabled'] = self.__dict__.pop('enabled')
        elif 'disabled' in self.__dict__ and 'disabled' not in kwargs:
            kwargs['disabled'] = self.__dict__.pop('disabled')
        self._update(**kwargs)
        self.__dict__['translationAddress'] = stash_translation_address



# coding=utf-8
# Copyright 2016 F5 Networks Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#

'''Class to manage a TrustDomain for a set of BIG-IP® devices.

A trust domain defines a group of devices that have signed and exchanged
certificates. Establishing a trust domain is prerequisite for device service
clustering. Once devices are part of a trust domain, they can synchronize
configuration and act as failovers for one another. This class manages that
trust domain.

Examples:

    devices = [ManagementRoot('x', 'un', 'pw'), ManagementRoot('x', 'un', 'pw')

    * Existing domain:
        dg = DeviceGroup(devices=devices, partition='Common')

    * New domain:
        dg = DeviceGroup()
        dg.create(devices=devices, partition='Common')

Methods:

    * validate -- ensure devices are a part of the same trust domain
    * create -- create a trust domain amongst two or more devices
    * teardown -- teardown a trust domain by traversing each member and
                  removing all other devices from its local trust besides
                  itself.

'''
from six import iteritems

from f5.multi_device.exceptions import DeviceAlreadyInTrustDomain
from f5.multi_device.exceptions import DeviceNotTrusted

from f5.multi_device.device_group import DeviceGroup
from f5.multi_device.utils import get_device_info
from f5.multi_device.utils import get_device_names_to_objects
from f5.multi_device.utils import pollster


class TrustDomain(object):
    '''Manages the trust domain of a BIG-IP® device.'''

    iapp_actions = {'definition': {'implementation': None, 'presentation': ''}}

    def __init__(self, **kwargs):
        '''Initialize a trusted peer manager object.

        The device_group_name set below is the default trust group that exists
        on all BIG-IP® devices. We are fixing it here to that group.

        :param kwargs: dict -- keyword args for devices and partition
        '''

        self.domain = {}
        if kwargs:
            self._set_attributes(**kwargs)
            self.validate()

    def _set_attributes(self, **kwargs):
        '''Set attributes for instance in one place

        :param kwargs: dict -- dictionary of keyword arguments
        '''

        self.devices = kwargs['devices'][:]
        self.partition = kwargs['partition']
        self.device_group_name = 'device_trust_group'
        self.device_group_type = 'sync-only'

    def validate(self):
        '''Validate that devices are each trusted by one another

        :param kwargs: dict -- keyword args for devices and partition
        :raises: DeviceNotTrusted
        '''

        self._populate_domain()
        missing = []
        for domain_device in self.domain:
            for truster, trustees in iteritems(self.domain):
                if domain_device not in trustees:
                    missing.append((domain_device, truster, trustees))

        if missing:
            msg = ''
            for item in missing:
                msg += '\n%r is not trusted by %r, which trusts: %r' % \
                    (item[0], item[1], item[2])
            raise DeviceNotTrusted(msg)
        self.device_group = DeviceGroup(
            devices=self.devices,
            device_group_name=self.device_group_name,
            device_group_type=self.device_group_type,
            device_group_partition=self.partition
        )

    def _populate_domain(self):
        '''Populate TrustDomain's domain attribute.

        This entails an inspection of each device's certificate-authority
        devices in its trust domain and recording them. After which, we
        get a dictionary of who trusts who in the domain.
        '''

        self.domain = {}
        for device in self.devices:
            device_name = get_device_info(device).name
            ca_devices = \
                device.tm.cm.trust_domains.trust_domain.load(
                    name='Root'
                ).caDevices
            self.domain[device_name] = [
                d.replace('/%s/' % self.partition, '') for d in ca_devices
            ]

    def create(self, **kwargs):
        '''Add trusted peers to the root bigip device.

        When adding a trusted device to a device, the trust is reflexive. That
        is, the truster trusts the trustee and the trustee trusts the truster.
        So we only need to add the trusted devices to one device.

        :param kwargs: dict -- devices and partition
        '''

        self._set_attributes(**kwargs)
        for device in self.devices[1:]:
            self._add_trustee(device)
        pollster(self.validate)()

    def teardown(self):
        '''Teardown trust domain by removing trusted devices.'''

        for device in self.devices:
            self._remove_trustee(device)
            self._populate_domain()
        self.domain = {}

    def _add_trustee(self, device):
        '''Add a single trusted device to the trust domain.

        :param device: ManagementRoot object -- device to add to trust domain
        '''

        device_name = get_device_info(device).name
        if device_name in self.domain:
            msg = 'Device: %r is already in this trust domain.' % device_name
            raise DeviceAlreadyInTrustDomain(msg)
        self._modify_trust(self.devices[0], self._get_add_trustee_cmd, device)

    def _remove_trustee(self, device):
        '''Remove a trustee from the trust domain.

        :param device: MangementRoot object -- device to remove
        '''

        trustee_name = get_device_info(device).name
        name_object_map = get_device_names_to_objects(self.devices)
        delete_func = self._get_delete_trustee_cmd

        for truster in self.domain:
            if trustee_name in self.domain[truster] and \
                    truster != trustee_name:
                truster_obj = name_object_map[truster]
                self._modify_trust(truster_obj, delete_func, trustee_name)
            self._populate_domain()

        for trustee in self.domain[trustee_name]:
            if trustee_name != trustee:
                self._modify_trust(device, delete_func, trustee)

        self.devices.remove(name_object_map[trustee_name])

    def _modify_trust(self, truster, mod_peer_func, trustee):
        '''Modify a trusted peer device by deploying an iapp.


        :param truster: ManagementRoot object -- device on which to perform
                        commands
        :param mod_peer_func: function -- function to call to modify peer
        :param trustee: ManagementRoot object or str -- device to modify
        '''

        iapp_name = 'trusted_device'
        mod_peer_cmd = mod_peer_func(trustee)
        iapp_actions = self.iapp_actions.copy()
        iapp_actions['definition']['implementation'] = mod_peer_cmd
        self._deploy_iapp(iapp_name, iapp_actions, truster)
        self._delete_iapp(iapp_name, truster)

    def _delete_iapp(self, iapp_name, deploying_device):
        '''Delete an iapp service and template on the root device.

        :param iapp_name: str -- name of iapp
        :param deploying_device: ManagementRoot object -- device where the
                                 iapp will be deleted
        '''

        iapp = deploying_device.tm.sys.application
        iapp_serv = iapp.services.service.load(
            name=iapp_name, partition=self.partition
        )
        iapp_serv.delete()
        iapp_tmpl = iapp.templates.template.load(
            name=iapp_name, partition=self.partition
        )
        iapp_tmpl.delete()

    def _deploy_iapp(self, iapp_name, actions, deploying_device):
        '''Deploy iapp to add trusted device

        :param iapp_name: str -- name of iapp
        :param actions: dict -- actions definition of iapp sections
        :param deploying_device: ManagementRoot object -- device where the
                                 iapp will be created
        '''

        tmpl = deploying_device.tm.sys.application.templates.template
        serv = deploying_device.tm.sys.application.services.service
        tmpl.create(name=iapp_name, partition=self.partition, actions=actions)
        pollster(deploying_device.tm.sys.application.templates.template.load)(
            name=iapp_name, partition=self.partition
        )
        serv.create(
            name=iapp_name,
            partition=self.partition,
            template='/%s/%s' % (self.partition, iapp_name)
        )

    def _get_add_trustee_cmd(self, trustee):
        '''Get tmsh command to add a trusted device.

        :param trustee: ManagementRoot object -- device to add as trusted
        :returns: str -- tmsh command to add trustee
        '''

        trustee_info = pollster(get_device_info)(trustee)
        username = trustee._meta_data['username']
        password = trustee._meta_data['password']
        return 'tmsh::modify cm trust-domain Root ca-devices add ' \
            '\\{ %s \\} name %s username %s password %s' % \
            (trustee_info.managementIp, trustee_info.name, username, password)

    def _get_delete_trustee_cmd(self, trustee_name):
        '''Get tmsh command to delete a trusted device.

        :param trustee_name: str -- name of device to remove
        :returns: str -- tmsh command to delete trusted device
        '''

        return 'tmsh::modify cm trust-domain Root ca-devices delete ' \
            '\\{ %s \\}' % trustee_name



# coding=utf-8
#
#  Copyright 2017 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


from f5.bigip.resource import Collection
from f5.bigip.resource import OrganizingCollection
from f5.bigip.resource import Resource
from f5.sdk_exception import ConstraintError
from f5.sdk_exception import MissingUpdateParameter


class Authz(OrganizingCollection):
    def __init__(self, shared):
        super(Authz, self).__init__(shared)
        self._meta_data['allowed_lazy_attributes'] = [
            Tokens_s
        ]


class Tokens_s(Collection):
    def __init__(self, authz):
        super(Tokens_s, self).__init__(authz)
        self._meta_data['required_json_kind'] = 'shared:authz:tokens:authtokencollectionstate'
        self._meta_data['allowed_lazy_attributes'] = [Token]
        self._meta_data['attribute_registry'] = {
            'shared:authz:tokens:authtokenitemstate': Token
        }


class Token(Resource):
    def __init__(self, tokens):
        super(Token, self).__init__(tokens)
        self._meta_data['required_json_kind'] = 'shared:authz:tokens:authtokenitemstate'

        # The required parameters are a little vague. It turns out that the "user"
        # value that is required is the "link" to the ID
        self._meta_data['required_creation_parameters'] = {'token', 'user'}

    def update(self, **kwargs):
        self._validate_params(**kwargs)
        return self._update(**kwargs)

    def modify(self, **kwargs):
        self._validate_timeout(**kwargs)
        return self._modify(**kwargs)

    def _validate_params(self, **kwargs):
        self._validate_timeout(**kwargs)
        self._validate_user(**kwargs)

    def _validate_user(self, **kwargs):
        try:
            assert 'user' in kwargs
            assert 'link' in kwargs['user']
        except AssertionError:
            raise MissingUpdateParameter(
                "The 'user' parameter is required when updating."
            )

    def _validate_timeout(self, **kwargs):
        timeout = kwargs.get('timeout', None)
        try:
            if timeout is not None and int(timeout) > 36000:
                raise ConstraintError(
                    "The provided timeout exceeds the limit of 36000."
                )
        except ValueError:
            raise ConstraintError(
                "The provided timeout must be a number between 1 and 36000."
            )



# Copyright 2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

'''Source engine container, combines code on filesystem with other sources.'''
from __future__ import print_function

LICENSE_AND_MODULE_DOCSTRING = """# Copyright 2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
import ast
import io
import os
from urlparse import urlparse


class UnexpectedNodeInFromImports(Exception):
    '''If a nonstring, nonimport is found before imports end raise this.'''
    pass


class SourceEngine(object):
    '''Given a Python source string, and a uri, integrate with modules

    Defines filesystem paths from the uri
    Checks for module based on path
    If no module dump string and done
    If module read in as string
    check class names, if they exist, skip
    Else:
        Handle import integration (additive)
    Append new class
    '''
    def __init__(self, source_root_dir, templ_src_str, selfLinkURI):
        # find/read source from module
        self.root_dir = source_root_dir
        self.module_path = self._build_path_from_URI(selfLinkURI)
        self.src_from_module = self._read_module_source(self.module_path)

        # parse_and_store module source
        self._parse_and_store_src('module', self.src_from_module)

        # parse and store template source
        self.src_from_template = templ_src_str
        self._parse_and_store_src('template', self.src_from_template)

    def _build_path_from_URI(self, selfLinkURI):
        _, _, uri_path, _, _, _ = urlparse(selfLinkURI)
        splits = uri_path.split('/')
        self.org_coll_dir = splits[3]
        if len(splits) > 4:
            self.module_name = splits[4] + '.py'
        else:
            self.module_name = '__init__.py'
        return os.path.join(self.root_dir, self.org_coll_dir, self.module_name)

    def _read_module_source(self, pathname):
        if os.path.isfile(pathname):
            with io.open(pathname, 'r') as fh:
                return fh.read()
        else:
            if not os.path.isdir(os.path.dirname(pathname)):
                os.makedirs(os.path.dirname(pathname))
            return ''

    def _get_ClassDefs(self, target_ast, found):
        for node in ast.walk(target_ast):
            if isinstance(node, ast.ClassDef):
                found[node.name] = node

    def _get_ImportFroms(self, target_ast, found):
        line_numbers = []
        for node in ast.walk(target_ast):
            if isinstance(node, ast.ImportFrom):
                line_numbers.append(node.lineno)
                fullname = 'from ' + node.module + ' import ' +\
                    node.names[0].name
                found[fullname] = node
        if not line_numbers:
            line_numbers.append(-1)
        found['start'] = min(line_numbers)
        found['end'] = max(line_numbers)
        # check for unexpected expressions in ImportFroms block
        for node in ast.walk(target_ast):
            if (getattr(node, 'lineno', False)):
                if node.lineno <= found['end']:
                    if (not isinstance(node, ast.ImportFrom)) and\
                       (not isinstance(node, ast.Str)) and\
                       (not isinstance(node, ast.Expr)):
                        raise UnexpectedNodeInFromImports(node)

    def _parse_and_store_src(self, src_type, src_string):
        print(src_string)
        setattr(self, 'src_from_'+src_type, src_string)
        setattr(self, src_type+'_ast', ast.parse(src_string))
        setattr(self, src_type+'_classdefs', {})
        self._get_ClassDefs(getattr(self, src_type+'_ast'),
                            getattr(self, src_type+'_classdefs'))
        setattr(self, src_type+'_fromimports', {})
        self._get_ImportFroms(getattr(self, src_type+'_ast'),
                              getattr(self, src_type+'_fromimports'))

    def _integrate_importfroms(self):
        start_ind = self.module_fromimports['start']
        end_ind = self.module_fromimports['end']
        fis_to_add = []
        for fromimport in self.template_fromimports:
            if fromimport not in self.module_fromimports:
                fis_to_add.append(fromimport)
        mod_fimport_lines = self.src_from_module_lines[start_ind-1:end_ind]
        mod_fimport_lines.extend(fis_to_add)
        mod_fimport_lines.sort()
        return mod_fimport_lines

    def _clean_template_string(self):
        ftl_raw = self.src_from_template.strip().splitlines()[
            self.template_fromimports['end']:]
        from_template_lines =\
            os.linesep.join(ftl_raw).strip().split(os.linesep)
        from_template_lines[-1] = from_template_lines[-1] + os.linesep
        from_template_lines.insert(0, os.linesep)
        return os.linesep.join(from_template_lines)

    def integrate(self):
        if not os.path.isfile(self.module_path):
            to_mod_string =\
                LICENSE_AND_MODULE_DOCSTRING + self.src_from_template
            with io.open(self.module_path, 'wb') as ofh:
                ofh.write(to_mod_string)
                return
        for key in self.template_classdefs:
            if key in self.module_classdefs:
                print('class: %r already defined!' % key)
                return
        self.src_from_module_lines = self.src_from_module.strip().splitlines()
        importfips_lines = self._integrate_importfroms()
        preamble_lines =\
            self.src_from_module_lines[:self.module_fromimports['start']-1]
        post_imp_lines =\
            self.src_from_module_lines[self.module_fromimports['end']:]
        post_imp_lines[-1] = post_imp_lines[-1]+os.linesep
        preamble_lines.extend(importfips_lines)
        preamble_lines.extend(post_imp_lines)
        module_src = os.linesep.join(preamble_lines)
        new_module_src = module_src + self._clean_template_string()
        os.rename(self.module_path, self.module_path+'.old')
        with io.open(self.module_path, 'wb') as ofh:
            ofh.write(new_module_src)



# Copyright 2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import re

from f5.sdk_exception import F5SDKError


class IappParser(object):

    template_sections = [
        'presentation',
        'implementation',
        'html-help',
        'role-acl'
    ]

    tcl_list_for_attr_re = r'{(\s*(\w+)?\s*)+}'
    tcl_list_for_section_re = r'(\s*\w+\s*)+'
    section_map = {
        'html-help': 'htmlHelp',
        'role-acl': 'roleAcl'
    }
    attr_map = {
        'requires-modules': 'requiresModules',
        'ignore-verification': 'ignoreVerification',
        'tmpl-signature': 'tmplSignature',
        'requires-bigip-version-min': 'requiresBigipVersionMin',
        'requires-bigip-version-max': 'requiresBigipVersionMax',
        'total-signing-status': 'totalSigningStatus',
        'prerequisite-errors': 'prerequisiteErrors',
        'verification-status': 'verificationStatus',
        'signing-key': 'signingKey',
        'tmpl-checksum': 'tmplChecksum'
    }
    sections_not_required = ['html-help', 'role-acl', 'macro']
    tcl_list_patterns = {
        'requires-modules': tcl_list_for_attr_re,
        'role-acl': tcl_list_for_section_re
    }
    template_attrs = [
        'description',
        'partition',
        'requires-modules',
        'ignore-verification',
        'requires-bigip-version-max',
        'requires-bigip-version-min',
        'signing-key',
        'tmpl-checksum',
        'tmpl-signature',
        'total-signing-status',
        'prerequisite-errors',
        'verification-status',
    ]

    def __init__(self, template_str):
        '''Initialize class.

        :param template_str: string of iapp template file
        :raises: EmptyTemplateException
        '''

        if template_str:
            self.template_str = str(template_str)
        else:
            raise EmptyTemplateException('Template empty or None value.')

    def _get_section_end_index(self, section, section_start):
        '''Get end of section's content.

        In the loop to match braces, we must not count curly braces that are
        within a doubly quoted string.

        :param section: string name of section
        :param section_start: integer index of section's beginning
        :return: integer index of section's end
        :raises: CurlyBraceMismatchException
        '''

        brace_count = 0
        in_quote = False
        in_escape = False

        for index, char in enumerate(self.template_str[section_start:]):
            # This check is to look for items inside of an escape sequence.
            #
            # For example, in the iApp team's iApps, there is a proc called
            # "iapp_get_items" which has a line that looks like this.
            #
            #     set val [string map {\" ""} $val]
            #
            # This will cause this parser to fail because of the unbalanced
            # quotes. Therefore, this conditional takes this into consideration
            #
            if char == '\\' and not in_escape:
                in_escape = True
            elif char == '\\' and in_escape:
                in_escape = False

            if not in_escape:
                if char == '"' and not in_quote:
                    in_quote = True
                elif char == '"' and in_quote:
                    in_quote = False

            if char == '{' and not in_quote:
                brace_count += 1
            elif char == '}' and not in_quote:
                brace_count -= 1

            if brace_count is 0:
                return index + section_start

        if brace_count is not 0:
            raise CurlyBraceMismatchException(
                'Curly braces mismatch in section %s.' % section
                )

    def _get_section_start_index(self, section):
        '''Get start of a section's content.

        :param section: string name of section
        :return: integer index of section's beginning
        :raises: NonextantSectionException
        '''

        sec_start_re = r'%s\s*\{' % section

        found = re.search(sec_start_re, self.template_str)
        if found:
            return found.end() - 1

        raise NonextantSectionException(
            'Section %s not found in template' % section
            )

    def _get_template_name(self):
        '''Find template name.

        :returns: string of template name
        :raises: NonextantTemplateNameException
        '''

        start_pattern = r"sys application template\s+" \
                        r"(\/[\w\.\-]+\/)?" \
                        r"(?P<name>[\w\.\-]+)\s*\{"

        template_start = re.search(start_pattern, self.template_str)
        if template_start:
            return template_start.group('name')

        raise NonextantTemplateNameException('Template name not found.')

    def _get_template_attr(self, attr):
        '''Find the attribute value for a specific attribute.

        :param attr: string of attribute name
        :returns: string of attribute value
        '''

        attr_re = r'{0}\s+.*'.format(attr)

        attr_found = re.search(attr_re, self.template_str)
        if attr_found:
            attr_value = attr_found.group(0).replace(attr, '', 1)
            return attr_value.strip()

    def _add_sections(self):
        '''Add the found and required sections to the templ_dict.'''
        for section in self.template_sections:
            try:
                sec_start = self._get_section_start_index(section)
            except NonextantSectionException:
                if section in self.sections_not_required:
                    continue
                raise
            sec_end = self._get_section_end_index(section, sec_start)
            section_value = self.template_str[sec_start+1:sec_end].strip()
            section, section_value = self._transform_key_value(
                section,
                section_value,
                self.section_map
            )
            self.templ_dict['actions']['definition'][section] = section_value
            self.template_str = self.template_str[:sec_start+1] + \
                self.template_str[sec_end:]

    def _add_cli_scripts(self):
        '''Add the found external sections to the templ_dict.'''

        pattern = r"cli script\s+" \
                  r"(\/[\w\.\-]+\/)?" \
                  r"(?P<name>[\w\.\-]+)\s*\{"
        sections = re.finditer(pattern, self.template_str)

        for section in sections:
            if 'scripts' not in self.templ_dict:
                self.templ_dict['scripts'] = []

            try:
                sec_start = self._get_section_start_index(
                    section.group('name')
                )
            except NonextantSectionException:
                continue

            sec_end = self._get_section_end_index(
                section.group('name'), sec_start
            )
            section_value = self.template_str[sec_start+1:sec_end].strip()

            self.templ_dict['scripts'].append(dict(
                name=section.group('name'),
                script=section_value
            ))

            self.template_str = self.template_str[:sec_start+1] + \
                self.template_str[sec_end:]

    def _add_attrs(self):
        '''Add the found and required attrs to the templ_dict.'''
        for attr in self.template_attrs:
            attr_value = self._get_template_attr(attr)

            if not attr_value:
                continue

            attr, attr_value = self._transform_key_value(
                attr,
                attr_value,
                self.attr_map
            )
            self.templ_dict[attr] = attr_value

    def _parse_tcl_list(self, attr, list_str):
        '''Turns a string representation of a TCL list into a Python list.

        :param attr: string name of attribute
        :param list_str: string representation of a list
        :returns: Python list
        '''

        list_str = list_str.strip()

        if not list_str:
            return []

        if list_str[0] != '{' and list_str[-1] != '}':
            if list_str.find('none') >= 0:
                return list_str

        if not re.search(self.tcl_list_patterns[attr], list_str):
            raise MalformedTCLListException(
                'TCL list for "%s" is malformed. ' % attr
            )

        list_str = list_str.strip('{').strip('}')
        list_str = list_str.strip()
        return list_str.split()

    def _transform_key_value(self, key, value, map_dict):
        '''Massage keys and values for iapp dict to look like JSON.

        :param key: string dictionary key
        :param value: string dictionary value
        :param map_dict: dictionary to map key names
        '''

        if key in self.tcl_list_patterns:
            value = self._parse_tcl_list(key, value)

        if key in map_dict:
            key = map_dict[key]

        return key, value

    def parse_template(self):
        '''Parse the template string into a dict.

        Find the (large) inner sections first, save them, and remove them from
        a modified string. Then find the template attributes in the modified
        string.

        :returns: dictionary of parsed template
        '''

        self.templ_dict = {'actions': {'definition': {}}}

        self.templ_dict['name'] = self._get_template_name()

        self._add_cli_scripts()
        self._add_sections()
        self._add_attrs()

        return self.templ_dict


class EmptyTemplateException(F5SDKError):
    pass


class CurlyBraceMismatchException(F5SDKError):
    pass


class NonextantSectionException(F5SDKError):
    pass


class NonextantTemplateNameException(F5SDKError):
    pass


class MalformedTCLListException(F5SDKError):
    pass



# coding=utf-8
#
#  Copyright 2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""BIG-IP® cluster traffic-group submodule

REST URI
    ``http://localhost/mgmt/tm/cm/traffic-group``

GUI Path
    ``Device Management --> Traffic Groups``

REST Kind
    ``tm:cm:traffic-group:*``
"""

from f5.bigip.resource import Collection
from f5.bigip.resource import Resource


class Traffic_Groups(Collection):
    """BIG-IP® cluster traffic-group collection"""
    def __init__(self, cm):
        super(Traffic_Groups, self).__init__(cm)
        self._meta_data['allowed_lazy_attributes'] = [Traffic_Group]
        self._meta_data['attribute_registry'] =\
            {'tm:cm:traffic-group:traffic-groupstate': Traffic_Group}


class Traffic_Group(Resource):
    """BIG-IP® cluster traffic-group resource"""
    def __init__(self, traffic_groups):
        super(Traffic_Group, self).__init__(traffic_groups)
        self._meta_data['required_json_kind'] =\
            'tm:cm:traffic-group:traffic-groupstate'
        self._meta_data['required_creation_parameters'].update(('partition',))



# coding=utf-8
#
# Copyright 2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""BIG-IP® system peformance stats module.

REST URI
    ``http://localhost/mgmt/tm/sys/performance``

GUI Path
    ``System --> Users --> Partition List``

REST Kind
    ``tm:sys:performance:*``
"""

from f5.bigip.resource import Collection
from f5.bigip.resource import UnnamedResource
from f5.sdk_exception import UnsupportedOperation


class Performances(Collection):
    """BIG-IP® system performace stats collection"""
    def __init__(self, sys):
        super(Performances, self).__init__(sys)
        self._meta_data['allowed_lazy_attributes'] = [All_Stats]

    def get_collection(self):
        '''Performance collections are not proper BIG-IP® collection objects.

        :raises: :exc:`~f5.BIG-IP.resource.UnsupportedOperation`
        '''
        raise UnsupportedOperation(
            "The iControl REST URI mgmt/sys/performance/ does not respond GET requests."
        )


class All_Stats(UnnamedResource):
    """BIG-IP® system performace stats unnamed resource"""
    def __init__(self, performance):
        super(All_Stats, self).__init__(performance)
        self._meta_data['required_load_parameters'] = set()
        self._meta_data['required_json_kind'] =\
            'tm:sys:performance:all-stats:all-statsstats'

    def update(self, **kwargs):
        '''Update is not supported for statistics.

        :raises: :exc:`~f5.BIG-IP.resource.UnsupportedOperation`
        '''
        raise UnsupportedOperation('Stats do not support create, only load and refresh')



# coding=utf-8
#
# Copyright 2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Directory: net module: fdb.

REST URI
    ``https://localhost/mgmt/tm/net/fdb``

GUI Path
    ``XXX``

REST Kind
    ``tm:net:fdb:*``
"""


from f5.bigip.resource import Collection
from f5.bigip.resource import OrganizingCollection
from f5.bigip.resource import Resource


class Fdb(OrganizingCollection):
    """BIG-IP® FDB collection."""
    def __init__(self, net):
        super(Fdb, self).__init__(net)
        self._meta_data['allowed_lazy_attributes'] = [Tunnels]
        # self._meta_data['icontrol_version'] = '11.5.0'


class Tunnel(Resource):
    """BIG-IP® Tunnel resource."""
    def __init__(self, Tunnels):
        super(Tunnel, self).__init__(Tunnels)
        self._meta_data['required_json_kind'] = "tm:net:fdb:tunnel:tunnelstate"
        self._meta_data['attribute_registry'] = {
            'tm:net:fdb:tunnel:records:recordscollectionstate': Records_s}
        # Setting this here to be explicit, even though it is set via
        # the super call from its containing object.
        # self._meta_data['icontrol_version'] = '11.5.0'


class Tunnels(Collection):
    """BIG-IP® Tunnels collection."""
    def __init__(self, fdb):
        super(Tunnels, self).__init__(fdb)
        self._meta_data['allowed_lazy_attributes'] = [Tunnel]
        self._meta_data['attribute_registry'] =\
            {'tm:net:fdb:tunnel:tunnelstate': Tunnel}
        # Setting this here to be explicit, even though it is set via
        # the super call from its containing object.
        # self._meta_data['icontrol_version'] = '11.5.0'


class Records_s(Collection):
    """BIG-IP® Tunnel records collection."""
    def __init__(self, tunnel):
        super(Records_s, self).__init__(tunnel)
        self._meta_data['allowed_lazy_attributes'] = [Records]
        self._meta_data['required_json_kind'] =\
            'tm:net:fdb:tunnel:records:recordscollectionstate'
        self._meta_data['attribute_registry'] =\
            {'tm:net:fdb:tunnel:records:recordsstate': Records}


class Records(Resource):
    """BIG-IP® LTM pool members sub-collection resource"""
    def __init__(self, records_s):
        super(Records, self).__init__(records_s)
        self._meta_data['required_json_kind'] =\
            'tm:net:fdb:tunnel:records:recordsstate'
        self._meta_data['required_creation_parameters'].update(('name',))
